from mmhuman3d.core.conventions.keypoints_mapping.human_data import (  # noqa:F401,E501
    APPROXIMATE_MAP, APPROXIMATE_MAPPING_LIST, HUMAN_DATA, HUMAN_DATA_BODY,
    HUMAN_DATA_HEAD, HUMAN_DATA_HIP, HUMAN_DATA_LEFT_HAND, HUMAN_DATA_LIMBS,
    HUMAN_DATA_LIMBS_INDEX, HUMAN_DATA_PALETTE, HUMAN_DATA_PARTS,
    HUMAN_DATA_RIGHT_HAND, HUMAN_DATA_SHOULDER,
)

HUMAN_DATA_LIMB_NAMES = {
    'left_ankle': {
        'left_knee': 'left_lower_leg'
    },
    'right_ankle': {
        'right_knee': 'right_lower_leg'
    },
    'left_shoulder': {
        'left_elbow': 'left_upperarm'
    },
    'right_shoulder': {
        'right_elbow': 'right_upperarm'
    },
    'left_elbow': {
        'left_wrist': 'left_forearm'
    },
    'right_elbow': {
        'right_wrist': 'right_forearm'
    },
    'left_hip_extra': {
        'left_knee': 'left_thigh'
    },
    'right_hip_extra': {
        'right_knee': 'right_thigh'
    },
}



from mmcv.utils import Registry

from .smpl import SMPL
from .smplx import SMPLX

BODYMODELS = Registry('body_model')

BODYMODELS.register_module(name='SMPL', module=SMPL)
BODYMODELS.register_module(name='SMPLX', module=SMPLX)


def build_body_model(cfg):
    """Build body model."""
    return BODYMODELS.build(cfg)



# yapf: disable

import logging
import math
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from abc import ABCMeta
from torch.nn.init import constant_, normal_, xavier_uniform_
from typing import Union
from xrprimer.utils.log_utils import get_logger

from xrmocap.model.loss.builder import build_loss
from xrmocap.model.mvp.builder import build_model
from xrmocap.model.mvp.position_encoding import get_2d_coords, get_rays
from xrmocap.model.mvp.projattn import ProjAttn
from xrmocap.utils.mvp_utils import absolute2norm, get_clones, inverse_sigmoid
from .base_architecture import BaseArchitecture

try:
    import Deformable as DF  # noqa F401
    has_deformable = True
    import_exception = ''
except (ImportError, ModuleNotFoundError):
    has_deformable = False
    import traceback
    import_exception = traceback.format_exc()

# yapf: enable


class MviewPoseTransformer(BaseArchitecture, metaclass=ABCMeta):
    """Multi-view Pose Transformer Module.

    Modified from DETR and Deformable Detr
    https://github.com/facebookresearch/detr
    https://github.com/fundamentalvision/Deformable-DETR.

    More details can be found on the website.
    https://github.com/sail-sg/mvp
    """

    def __init__(self, is_train: bool, logger: Union[None, str,
                                                     logging.Logger],
                 n_kps: int, n_instance: int, image_size: list, d_model: int,
                 use_feat_level: list, n_cameras: int, query_embed_type: str,
                 with_pose_refine: bool, loss_weight_loss_ce: float,
                 loss_per_kp: float, aux_loss: bool, pred_conf_threshold: list,
                 pred_class_fuse: str, projattn_pos_embed_mode: str,
                 query_adaptation: bool, convert_kp_format_indexes: list,
                 backbone_setup: dict, proj_attn_setup: dict,
                 decoder_layer_setup: dict, decoder_setup: dict,
                 pos_encoding_setup: dict, pose_embed_setup: dict,
                 matcher_setup: dict, criterion_setup: dict, space_size: list,
                 space_center: list) -> None:
        """
        Args:
            is_train (bool):
                True if it is initialized during training.
            logger (Union[None, str, logging.Logger]):
                Logger for logging. If None, root logger will be selected.
            n_kps (int):
                Number of keypoints per person.
            n_instance (int):
                Max number of person the model can handle.
            image_size (list):
                Input image size.
            d_model (int):
                Size of model and feature size.
            use_feat_level (list):
                Index of backbone features used.
            n_cameras (int):
                Number of cameras.
            query_embed_type (str):
                Type of query embedding.
                ['person_kp','image_person_kp','per_kp'].
            with_pose_refine (bool):
                Whether to use pose refine.
            loss_weight_loss_ce (float):
                Loss weight for CE loss.
            loss_per_kp (float):
                Loss weight for KP loss.
            aux_loss (bool):
                Whether to split loss by decoder layers.
            pred_conf_threshold (list):
                List of confidence threshold to filter non-human keypoints.
            pred_class_fuse (str):
                Type of fusing predictions.
                ['mean', 'feat_mean_pool', 'feat_max_pool']
            projattn_pos_embed_mode (str):
                The positional embedding mode of projective
                attention. ['use_rayconv','use_2d_coordconv']
                query_adaptation (bool): Whether to use query adaptation.
            convert_kp_format_indexes (list):
                Convention from CMU panoptic keypoint format.
            backbone_setup (dict):
                Dict if parameters to setup the backbone.
            decoder_layer_setup (dict):
                Dict if parameters to setup decoder layers.
            decoder_setup (dict):
                Dict if parameters to setup the decoder.
            pos_encoding_setup (dict):
                Dict if parameters to setup positional encoding.
            pose_embed_setup (dict):
                Dict if parameters to setup pose embedding.
            matcher_setup (dict):
                Dict if parameters to setup the matcher.
            criterion_setup (dict):
                Dict if parameters to setup criterions.
            space_size (list):
                Size of the 3D space.
            space_center (list):
                Center position of the 3D space.

        """

        super(MviewPoseTransformer, self).__init__()
        self.logger = get_logger(logger)
        if not has_deformable:
            self.logger.error(import_exception)
            raise ModuleNotFoundError('Please install deformable to run MVP.')
        self.n_kps = n_kps
        self.n_instance = n_instance
        self.image_size = np.array(image_size)
        self.grid_size = torch.tensor(space_size)
        self.grid_center = torch.tensor(space_center)
        self.use_feat_level = use_feat_level

        self.reference_points = nn.Linear(d_model, 3)
        self.reference_feats = nn.Linear(d_model * len(self.use_feat_level) *
                                         n_cameras,
                                         d_model)  # 256*feat_level*num_views

        self.decoder_layer_setup = decoder_layer_setup

        backbone_cfg = dict(
            type='PoseResNet',
            logger=self.logger,
            inplanes=2048,
        )
        backbone_cfg.update(backbone_setup)
        self.backbone = build_model(backbone_cfg)

        if is_train:
            self.backbone.init_weights()

        # projective attention
        proj_attn_cfg = dict(type='ProjAttn', logger=self.logger)
        proj_attn_cfg.update(proj_attn_setup)
        proj_attn = build_model(proj_attn_cfg)

        decoder_layer_cfg = dict(type='MvPDecoderLayer', proj_attn=proj_attn)
        decoder_layer_cfg.update(decoder_layer_setup)
        decoder_layer = build_model(decoder_layer_cfg)

        decoder_cfg = dict(type='MvPDecoder', decoder_layer=decoder_layer)
        decoder_cfg.update(decoder_setup)
        self.decoder = build_model(decoder_cfg)

        n_queries = n_instance * n_kps

        self.query_embed_type = query_embed_type
        if self.query_embed_type == 'person_kp':
            self.joint_embedding = nn.Embedding(n_kps, d_model * 2)
            self.instance_embedding = nn.Embedding(n_instance, d_model * 2)
        elif self.query_embed_type == 'image_person_kp':
            self.image_embedding = nn.Embedding(1, d_model * 2)
            self.joint_embedding = nn.Embedding(n_kps, d_model * 2)
            self.instance_embedding = nn.Embedding(n_instance, d_model * 2)
        elif self.query_embed_type == 'per_kp':
            self.query_embed = nn.Embedding(n_queries, d_model * 2)

        n_steps = d_model // 2

        pos_encoding_cfg = dict(
            type='PositionEmbeddingSine', n_pos_feats=n_steps, scale=None)
        pos_encoding_cfg.update(pos_encoding_setup)
        self.pos_encoding = build_model(pos_encoding_cfg)

        self.view_embed = nn.Parameter(torch.Tensor(n_cameras, d_model))
        self._reset_parameters()

        # We can use gt camera for projection,
        # so we dont need to regress camera param
        n_pred = self.decoder.n_layers
        n_classes = 2

        pose_embed_cfg = dict(
            type='MLP',
            input_dim=pose_embed_setup.d_model,
            hidden_dim=pose_embed_setup.d_model,
            output_dim=3,
            n_layers=pose_embed_setup.pose_embed_layer)

        self.pose_embed = build_model(pose_embed_cfg)
        self.class_embed = nn.Linear(d_model, n_classes)

        prior_prob = 0.01
        bias_value = -math.log((1 - prior_prob) / prior_prob)
        self.class_embed.bias.data = torch.ones(n_classes) * bias_value
        nn.init.constant_(self.pose_embed.layers[-1].weight.data, 0)
        nn.init.constant_(self.pose_embed.layers[-1].bias.data, 0)

        if with_pose_refine:
            self.class_embed = get_clones(self.class_embed, n_pred)
            self.pose_embed = get_clones(self.pose_embed, n_pred)
            self.decoder.pose_embed = self.pose_embed
        else:
            nn.init.constant_(self.pose_embed.layers[-1].bias.data[2:], -2.0)
            self.class_embed = nn.ModuleList(
                [self.class_embed for _ in range(n_pred)])
            self.pose_embed = nn.ModuleList(
                [self.pose_embed for _ in range(n_pred)])
            self.decoder.pose_embed = None

        matcher_cfg = dict(
            type='HungarianMatcher', cost_class=2., cost_pose=5.)
        matcher_cfg.update(matcher_setup)
        matcher = build_model(matcher_cfg)

        weight_dict = {
            'loss_ce': loss_weight_loss_ce,
            'loss_per_kp': loss_per_kp,
        }

        losses = ['kps', 'labels', 'cardinality']
        self.aux_loss = aux_loss

        if self.aux_loss:
            aux_weight_dict = {}
            for i in range(n_pred - 1):
                aux_weight_dict.update(
                    {k + f'_{i}': v
                     for k, v in weight_dict.items()})
            aux_weight_dict.update(
                {k + f'{"_enc"}': v
                 for k, v in weight_dict.items()})
            weight_dict.update(aux_weight_dict)

        criterion_cfg = dict(
            type='SetCriterion',
            losses=losses,
            n_classes=n_classes,
            matcher=matcher,
            weight_dict=weight_dict)
        criterion_cfg.update(criterion_setup)
        self.criterion = build_loss(criterion_cfg)

        self.pred_conf_threshold = pred_conf_threshold
        self.pred_class_fuse = pred_class_fuse

        self.level_embed = nn.Parameter(torch.Tensor(3, d_model))
        self.projattn_pos_embed_mode = projattn_pos_embed_mode
        self.query_adaptation = query_adaptation
        self.convert_kp_format_indexes = convert_kp_format_indexes

    def get_valid_ratio(self, mask):
        _, H, W = mask.shape
        valid_H = torch.sum(~mask[:, :, 0], 1)
        valid_W = torch.sum(~mask[:, 0, :], 1)
        valid_ratio_h = valid_H.float() / H
        valid_ratio_w = valid_W.float() / W
        valid_ratio = torch.stack([valid_ratio_w, valid_ratio_h], -1)
        return valid_ratio

    def _reset_parameters(self):
        for m in self.modules():
            if isinstance(m, ProjAttn):
                m._reset_parameters()
        xavier_uniform_(self.reference_points.weight.data, gain=1.0)
        constant_(self.reference_points.bias.data, 0.)
        normal_(self.view_embed)

    @torch.jit.unused
    def _set_aux_loss(self, outputs_class, outputs_coord):
        return [{
            'pred_logits': a,
            'pred_poses': b
        } for a, b in zip(outputs_class[:-1], outputs_coord[:-1])]

    def collate_first_two_dims(self, tensor):
        dim0 = tensor.shape[0]
        dim1 = tensor.shape[1]
        left = tensor.shape[2:]
        return tensor.view(dim0 * dim1, *left)

    def forward(self, views=None, meta=None):
        if views is not None:
            all_feats = self.backbone(
                torch.cat(views, dim=0), self.use_feat_level)
            all_feats = all_feats[::-1]
        batch, _, imageh, imagew = views[0].shape
        n_view = len(views)

        cam_R = torch.stack([m['camera']['R'] for m in meta], dim=1)
        cam_T = torch.stack([m['camera']['camera_standard_T'] for m in meta],
                            dim=1)
        cam_K = torch.stack([m['camera']['K'] for m in meta], dim=1)

        affine_trans = torch.stack([m['affine_trans'] for m in meta], dim=1)
        cam_K_crop = \
            torch.bmm(
                self.collate_first_two_dims(affine_trans),
                self.collate_first_two_dims(cam_K)).view(batch, n_view, 3, 3)
        n_feat_level = len(all_feats)
        camera_rays = []
        # get pos embed, camera ray or 2d coords
        for lvl in range(n_feat_level):
            # this can be compute only once, without iterating over views
            if self.projattn_pos_embed_mode == 'use_rayconv':
                camera_rays.append(
                    get_rays(self.image_size, all_feats[lvl].shape[2],
                             all_feats[lvl].shape[3], cam_K_crop, cam_R,
                             cam_T).flatten(0, 1))
            elif self.projattn_pos_embed_mode == 'use_2d_coordconv':
                camera_rays.append(
                    get_2d_coords(self.image_size, all_feats[lvl].shape[2],
                                  all_feats[lvl].shape[3], cam_K_crop, cam_R,
                                  cam_T).flatten(0, 1))

        src_flatten_views = []
        mask_flatten_views = []
        spatial_shapes_views = []

        for lvl, src in enumerate(all_feats):
            bs, c, h, w = src.shape
            spatial_shape = (h, w)
            spatial_shapes_views.append(spatial_shape)
            mask = src.new_zeros(bs, h, w).bool()
            mask_flatten_views.append(mask)
            mask = mask.flatten(1)
            src_flatten_views.append(src)

        spatial_shapes_views = \
            torch.as_tensor(spatial_shapes_views,
                            dtype=torch.long,
                            device=mask.device)
        level_start_index_views = \
            torch.cat((mask.new_zeros((1, ), dtype=torch.long),
                       torch.as_tensor(spatial_shapes_views,
                                       dtype=torch.long,
                                       device=mask.device)
                       .prod(1).cumsum(0)[:-1]))
        valid_ratios_views = torch.stack(
            [self.get_valid_ratio(m) for m in mask_flatten_views], 1)
        mask_flatten_views = [m.flatten(1) for m in mask_flatten_views]

        # query embedding scheme
        if self.query_embed_type == 'person_kp':
            # person embedding + kp embedding
            kp_embeds = self.joint_embedding.weight.unsqueeze(0)
            instance_embeds = self.instance_embedding.weight.unsqueeze(1)
            query_embeds = (kp_embeds + instance_embeds).flatten(0, 1)

        if self.query_embed_type == 'image_person_kp':
            # image_embedding + person embedding + kp embedding
            kp_embeds = self.joint_embedding.weight.unsqueeze(0)
            instance_embeds = self.instance_embedding.weight.unsqueeze(1)
            query_embeds = (kp_embeds + instance_embeds).flatten(0, 1)
            query_embeds += self.image_embedding.weight

        elif self.query_embed_type == 'per_kp':
            # per kp embedding
            query_embeds = self.query_embed.weight

        query_embed, target = torch.split(query_embeds, c, dim=1)
        query_embed = query_embed.unsqueeze(0).expand(batch, -1, -1)
        target = target.unsqueeze(0).expand(batch, -1, -1)

        # query adaptation
        if self.query_adaptation:
            feats_0 = F.adaptive_avg_pool2d(all_feats[0], (1, 1))
            feats_1 = F.adaptive_avg_pool2d(all_feats[1], (1, 1))
            feats_2 = F.adaptive_avg_pool2d(all_feats[2], (1, 1))
            feats = torch.cat((feats_0, feats_1, feats_2),
                              dim=1).squeeze().view(1, -1)
            ref_feats = self.reference_feats(feats).unsqueeze(0)
            reference_points = self.reference_points(query_embed +
                                                     ref_feats).sigmoid()
        else:
            reference_points = self.reference_points(query_embed).sigmoid()

        init_reference = reference_points  # B x 150 x 3

        kps_features, inter_references = \
            self.decoder(target, reference_points, src_flatten_views,
                         camera_rays,
                         meta=meta, src_spatial_shapes=spatial_shapes_views,
                         src_level_start_index=level_start_index_views,
                         src_valid_ratios=valid_ratios_views,
                         query_pos=query_embed,
                         src_padding_mask=mask_flatten_views)
        outputs_classes = []
        outputs_coords = []
        for lvl in range(kps_features.shape[0]):
            if lvl == 0:
                reference = init_reference
            else:
                reference = inter_references[lvl - 1]
            reference = inverse_sigmoid(reference)
            # mean after sigmoid
            if self.pred_class_fuse == 'mean':
                outputs_class = self.class_embed[lvl](kps_features[lvl]).\
                    view(batch, self.n_instance, self.n_kps, -1).\
                    sigmoid().mean(2)
                outputs_class = inverse_sigmoid(outputs_class)
            elif self.pred_class_fuse == 'feat_mean_pool':
                outputs_class = self.class_embed[lvl](kps_features[lvl])\
                    .view(batch, self.n_instance, self.n_kps, -1)\
                    .mean(2)
            elif self.pred_class_fuse == 'feat_max_pool':
                outputs_class = \
                    self.class_embed[lvl](
                        kps_features[lvl].view(batch,
                                               self.n_instance,
                                               self.n_kps, -1).max(2)[0])
            else:
                raise NotImplementedError
            tmp = self.pose_embed[lvl](kps_features[lvl])
            tmp += reference
            outputs_coord = tmp.sigmoid()

            outputs_classes.append(outputs_class)

            # convert panoptic kps to shelf/campus
            if self.convert_kp_format_indexes is not None:
                outputs_coord = \
                    outputs_coord.view(batch,
                                       self.n_instance,
                                       self.n_kps, -1)
                outputs_coord \
                    = outputs_coord[..., self.convert_kp_format_indexes, :]
                outputs_coord = outputs_coord.flatten(1, 2)

            outputs_coords.append({'outputs_coord': outputs_coord})

        out = {
            'pred_logits': outputs_classes[-1],
            'pred_poses': outputs_coords[-1]
        }

        if self.aux_loss:
            out['aux_outputs'] = \
                self._set_aux_loss(outputs_classes, outputs_coords)

        if self.training and 'kps3d' in meta[0] \
                and 'kps3d_vis' in meta[0]:
            meta[0]['kps3d_norm'] = \
                absolute2norm(meta[0]['kps3d'].float(),
                              self.grid_size,
                              self.grid_center)
            loss_dict, loss_value = self.criterion(out, meta)
            return out, loss_dict, loss_value

        return out

    def forward_train(self, **kwargs):
        """Forward train function for general training.

        For multi_view_pose transformer estimation, we do not use this
        interface.
        """
        raise NotImplementedError('This interface should not be used in '
                                  'current training schedule.')

    def forward_test(self, x: torch.Tensor, **kwargs):
        """Forward test function for general training.

        For multi_view_pose transformer estimation, we do not use this
        interface.
        """
        raise NotImplementedError('This interface should not be used in '
                                  'current training schedule.')



import torch

from .smplifyx import SMPLifyX


class SMPLifyXD(SMPLifyX):
    """Re-implementation of SMPLify-X with displacement."""
    OPTIM_PARAM = SMPLifyX.OPTIM_PARAM + [
        'displacement',
    ]

    def __prepare_optimizable_parameters__(self, init_dict: dict,
                                           batch_size: int) -> dict:
        """Prepare optimizable parameters in batch for registrant. If some of
        the parameters can be found in init_dict, use them for initialization.

        Args:
            init_dict (dict):
                A dict of init parameters. init_dict.keys() is a
                sub-set of self.__class__.OPTIM_PARAM.
            batch_size (int)

        Returns:
            dict:
                A dict of optimizable parameters, whose keys are
                self.__class__.OPTIM_PARAM and values are
                Tensors in batch.
        """
        smplx_init_dict = init_dict.copy()
        init_displacement = smplx_init_dict.pop('displacement', None)
        OPTIM_PARAM_backup = self.__class__.OPTIM_PARAM
        self.__class__.OPTIM_PARAM = SMPLifyX.OPTIM_PARAM
        ret_dict = SMPLifyX.__prepare_optimizable_parameters__(
            self, init_dict=smplx_init_dict, batch_size=batch_size)
        self.__class__.OPTIM_PARAM = OPTIM_PARAM_backup
        default_displacement = torch.zeros(
            size=(1, self.body_model.NUM_VERTS, 3),
            dtype=self.body_model.betas.dtype,
            device=self.device,
            requires_grad=True)
        displacement = self.__match_init_batch_size__(
            init_param=init_displacement,
            default_param=default_displacement,
            batch_size=batch_size)
        ret_dict['displacement'] = displacement
        return ret_dict



import torch
from typing import List


class OptimizableParameters():

    def __init__(self):
        """Collects parameters for optimization."""
        self.opt_params = {}

    def add_param(self,
                  key: str,
                  param: torch.Tensor,
                  fit_param: bool = True) -> None:
        """"Set requires_grad, collect the parameter for optimization.

        Args:
            key (str):
                Key of the param.
            param (torch.Tensor):
                Model parameter.
            fit_param (bool):
                Whether to optimize this body model parameter.
                Defaults to True.
        """
        if fit_param:
            param.requires_grad = True
        else:
            param.requires_grad = False
        self.opt_params[key] = param

    def set_param(self, key: str, fit_param: bool = True) -> None:
        """"Set requires_grad of a param in self.opt_params.

        Args:
            key (str):
                Key of the param.
            fit_param (bool):
                Whether to optimize this body model parameter.
                Defaults to True.
        """
        if fit_param:
            self.opt_params[key].requires_grad = True
        else:
            self.opt_params[key].requires_grad = False

    def parameters(self) -> List[torch.Tensor]:
        """Returns all parameters recorded by self. Compatible with mmcv's
        build_parameters()

        Returns:
            List[torch.Tensor]:
                A list of body model parameters for optimization.
        """
        ret_list = []
        for _, value in self.opt_params.items():
            ret_list.append(value)
        return ret_list



import logging
import numpy as np
from PIL import Image
from typing import Union

from .base_image_transform import BaseImageTransform


class CV2ToPIL(BaseImageTransform):

    def __init__(self,
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """Convert cv2 image array to PIL.Image.

        Args:
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """
        BaseImageTransform.__init__(self, logger=logger)

    def forward(self, input: np.ndarray) -> Image:
        """Forward function of CV2ToPIL.

        Args:
            input (np.ndarray):
                Image array defined in cv2.

        Returns:
            Image:
                Image instance defined in PIL.
        """
        pil_img = Image.fromarray(input)
        return pil_img



import numpy as np
import torch
from typing import List, Tuple

from xrmocap.data.dataset.base_dataset import BaseDataset


class MVPEvaluation:
    """Evaluation for MvP method."""

    def __init__(self, dataset: BaseDataset, kps_thr: float = 0.1):
        self.dataset = dataset
        self.gt_num = dataset.len
        self.n_views = dataset.n_views
        self.kps_thr = kps_thr

    def evaluate_map(self, pred_kps3d: torch.Tensor) \
            -> Tuple[List[float], List[float], float, float]:
        """Evaluate MPJPE, mAP and recall based on MPJPE. Mainly for panoptic
        predictions.

        Args:
            pred_kps3d (torch.Tensor):
                Predicted 3D keypoints.

        Returns:
            Tuple[List[float], List[float], float, float]:
                List of AP, list of recall, MPJPE value and recall@500mm.
        """

        eval_list = []
        assert len(pred_kps3d) == self.gt_num, \
            f'number mismatch {len(pred_kps3d)} pred and {self.gt_num} gt'

        trans_ground = torch.tensor([[1.0, 0.0, 0.0], [0.0, 0.0, -1.0],
                                     [0.0, 1.0, 0.0]]).double()

        total_gt = 0
        for index in range(self.gt_num):
            scene_idx, frame_idx, _ = self.dataset.process_index_mapping(index)
            gt_keypoints3d = self.dataset.gt3d[scene_idx]
            gt_scene_kps3d = gt_keypoints3d.get_keypoints(
            )  # [n_frame, n_person, n_kps, 4]
            gt_frame_kps3d = gt_scene_kps3d[
                frame_idx][:, :self.dataset.n_kps, :]  # [n_person, n_kps, 4]

            check_valid = torch.sum(gt_frame_kps3d, axis=1)  # [n_person, 4]
            gt_frame_kps3d = gt_frame_kps3d[check_valid[:, -1] > 0]

            gt_frame_kps3d = [
                torch.mm(gt_person_kps3d[:, 0:3], trans_ground)
                for gt_person_kps3d in gt_frame_kps3d
            ]
            if len(gt_frame_kps3d) == 0:
                continue

            pred_frame_kps3d = pred_kps3d[index].copy()
            pred_frame_kps3d_valid = pred_frame_kps3d[pred_frame_kps3d[:, 0,
                                                                       3] >= 0]

            for pred_person_kps3d in pred_frame_kps3d_valid:
                mpjpes = []
                for gt_person_kps3d in gt_frame_kps3d:
                    vis = gt_person_kps3d[:, -1] > self.kps_thr

                    mpjpe = np.mean(
                        np.sqrt(
                            np.sum(
                                (np.array(pred_person_kps3d[vis, 0:3]) -
                                 np.array(gt_person_kps3d[vis, 0:3]))**2,
                                axis=-1)))
                    mpjpes.append(mpjpe)
                min_gt = np.argmin(mpjpes)
                min_mpjpe = np.min(mpjpes)
                score = pred_person_kps3d[0, 4]
                eval_list.append({
                    'mpjpe': float(min_mpjpe),
                    'score': float(score),
                    'gt_id': int(total_gt + min_gt)
                })

            total_gt += len(gt_frame_kps3d)

        mpjpe_threshold = np.arange(25, 155, 25)
        aps = []
        recs = []
        for t in mpjpe_threshold:
            ap, rec = self._eval_list_to_ap(eval_list, total_gt, t)
            aps.append(ap)
            recs.append(rec)

        return \
            aps, \
            recs, \
            self._eval_list_to_mpjpe(eval_list), \
            self._eval_list_to_recall(eval_list, total_gt)

    def evaluate_pcp(self,
                     pred_kps3d: torch.Tensor,
                     recall_threshold: int = 500,
                     alpha: float = 0.5) -> Tuple[List[float], float, float]:
        """Evaluate MPJPE and PCP. Mainly for Shelf and Campus predictions.

        Args:
            pred_kps3d (torch.Tensor):
                Predicted 3D keypoints.
            recall_threshold (int, optional):
                Threshold for MPJPE. Defaults to 500.
            alpha (float, optional):
                Threshold for correct limb part. Defaults to 0.5.
                Predicted limb part is regarded as correct if
                predicted_part_length < alpha * gt_part_length
        Returns:
            Tuple[List[float], float, float]:
                List of PCP per actor, average PCP, and recall@500mm.
        """
        assert len(pred_kps3d) == self.gt_num, \
            f'number mismatch {len(pred_kps3d)} pred and {self.gt_num} gt'

        trans_ground = torch.tensor([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0],
                                     [0.0, 0.0, 1.0]]).double()

        # TODO: can get this info from Keypoints?
        limbs = [[0, 1], [1, 2], [3, 4], [4, 5], [6, 7], [7, 8], [9, 10],
                 [10, 11], [12, 13]]

        gt_n_person = np.max(
            np.array([
                scene_keypoints.get_keypoints().shape[1]
                for scene_keypoints in self.dataset.gt3d
            ]))

        correct_parts = np.zeros(gt_n_person)
        total_parts = np.zeros(gt_n_person)
        limb_correct_parts = np.zeros((gt_n_person, 10))

        total_gt = 0
        match_gt = 0
        for index in range(self.gt_num):
            scene_idx, frame_idx, _ = self.dataset.process_index_mapping(index)
            gt_keypoints3d = self.dataset.gt3d[scene_idx]
            gt_scene_kps3d = gt_keypoints3d.get_keypoints(
            )  # [n_frame, n_person, n_kps, 4]
            gt_frame_kps3d = gt_scene_kps3d[
                frame_idx][:, :self.dataset.n_kps, :]  # [n_person, n_kps, 4]

            gt_frame_kps3d = [
                torch.mm(gt_person_kps3d[:, 0:3], trans_ground)
                for gt_person_kps3d in gt_frame_kps3d
            ]
            if len(gt_frame_kps3d) == 0:
                continue

            pred_frame_kps3d = pred_kps3d[index].copy()
            pred_frame_kps3d_valid = pred_frame_kps3d[pred_frame_kps3d[:, 0, 3]
                                                      >= 0]  # if is a person

            for person_idx, gt_person_kps3d in enumerate(gt_frame_kps3d):

                vis = gt_person_kps3d[:, -1] > self.kps_thr

                check_valid = torch.sum(gt_person_kps3d, axis=1)  # [4]
                if check_valid[-1] == 0:
                    continue

                mpjpes = np.mean(
                    np.sqrt(
                        np.sum(
                            (np.array(pred_frame_kps3d_valid[:, vis, 0:3]) -
                             np.array(gt_person_kps3d[vis, 0:3]))**2,
                            axis=-1)),
                    axis=-1)
                min_n = np.argmin(mpjpes)
                min_mpjpe = np.min(mpjpes)

                if min_mpjpe < recall_threshold:
                    match_gt += 1

                total_gt += 1

                for j, k in enumerate(limbs):
                    total_parts[person_idx] += 1
                    error_s = \
                        np.linalg.norm(np.array(
                            pred_frame_kps3d_valid[min_n, k[0], 0:3]) -
                            np.array(gt_person_kps3d[k[0]]))
                    error_e = \
                        np.linalg.norm(np.array(
                            pred_frame_kps3d_valid[min_n, k[1], 0:3]) -
                            np.array(gt_person_kps3d[k[1]]))
                    limb_length = np.linalg.norm(
                        np.array(gt_person_kps3d[k[0]]) -
                        np.array(gt_person_kps3d[k[1]]))

                    if (error_s + error_e) / 2.0 <= alpha * limb_length:
                        correct_parts[person_idx] += 1
                        limb_correct_parts[person_idx, j] += 1

                # TODO: get these kps inedex from kps name
                pred_hip = (pred_frame_kps3d_valid[min_n, 2, 0:3] +
                            pred_frame_kps3d_valid[min_n, 3, 0:3]) / 2.0
                gt_hip = (gt_person_kps3d[2] + gt_person_kps3d[3]) / 2.0
                total_parts[person_idx] += 1
                error_s = np.linalg.norm(np.array(pred_hip) - np.array(gt_hip))
                error_e = np.linalg.norm(
                    np.array(pred_frame_kps3d_valid[min_n, 12, 0:3]) -
                    np.array(gt_person_kps3d[12]))
                limb_length = np.linalg.norm(
                    np.array(gt_hip) - np.array(gt_person_kps3d[12]))

                if (error_s + error_e) / 2.0 <= alpha * limb_length:
                    correct_parts[person_idx] += 1
                    limb_correct_parts[person_idx, 9] += 1

        actor_pcp = correct_parts / (total_parts + 1e-8)
        avg_pcp = np.mean(actor_pcp[:3])

        return \
            actor_pcp, avg_pcp, match_gt / (total_gt + 1e-8)

    def _eval_list_to_ap(self, eval_list, total_gt, threshold):
        eval_list.sort(key=lambda k: k['score'], reverse=True)
        total_num = len(eval_list)

        tp = np.zeros(total_num)
        fp = np.zeros(total_num)
        gt_det = []
        for i, item in enumerate(eval_list):
            if item['mpjpe'] < threshold and item['gt_id'] not in gt_det:
                tp[i] = 1
                gt_det.append(item['gt_id'])
            else:
                fp[i] = 1
        tp = np.cumsum(tp)
        fp = np.cumsum(fp)
        recall = tp / (total_gt + 1e-5)
        precise = tp / (tp + fp + 1e-5)
        for n in range(total_num - 2, -1, -1):
            precise[n] = max(precise[n], precise[n + 1])

        precise = np.concatenate(([0], precise, [0]))
        recall = np.concatenate(([0], recall, [1]))
        index = np.where(recall[1:] != recall[:-1])[0]
        ap = np.sum((recall[index + 1] - recall[index]) * precise[index + 1])

        return ap, recall[-2]

    def _eval_list_to_mpjpe(self, eval_list, threshold=500):
        eval_list.sort(key=lambda k: k['score'], reverse=True)
        gt_det = []

        mpjpes = []
        for i, item in enumerate(eval_list):
            if item['mpjpe'] < threshold and item['gt_id'] not in gt_det:
                mpjpes.append(item['mpjpe'])
                gt_det.append(item['gt_id'])

        return np.mean(mpjpes) if len(mpjpes) > 0 else np.inf

    def _eval_list_to_recall(self, eval_list, total_gt, threshold=500):
        gt_ids = [e['gt_id'] for e in eval_list if e['mpjpe'] < threshold]

        return len(np.unique(gt_ids)) / total_gt



# yapf: disable
import numpy as np
from typing import List, Tuple, Union, overload

from xrmocap.data_structure.keypoints import Keypoints

try:
    from mmhuman3d.core.visualization.visualize_keypoints2d import (
        visualize_kp2d,
    )
    has_mmhuman3d = True
    import_exception = ''
except (ImportError, ModuleNotFoundError):
    has_mmhuman3d = False
    import traceback
    stack_str = ''
    for line in traceback.format_stack():
        if 'frozen' not in line:
            stack_str += line + '\n'
    import_exception = traceback.format_exc() + '\n'
    import_exception = stack_str + import_exception
# yapf: enable


@overload
def visualize_keypoints2d(
        keypoints: Keypoints,
        output_path: str,
        img_arr: np.ndarray,
        overwrite: bool = False,
        return_array: bool = False) -> Union[None, np.ndarray]:
    ...


@overload
def visualize_keypoints2d(
        keypoints: Keypoints,
        output_path: str,
        img_paths: List[str],
        overwrite: bool = False,
        return_array: bool = False) -> Union[None, np.ndarray]:
    ...


@overload
def visualize_keypoints2d(
        keypoints: Keypoints,
        output_path: str,
        video_path: str,
        overwrite: bool = False,
        return_array: bool = False) -> Union[None, np.ndarray]:
    ...


@overload
def visualize_keypoints2d(
        keypoints: Keypoints,
        output_path: str,
        overwrite: bool = False,
        return_array: bool = False) -> Union[None, np.ndarray]:
    ...


def visualize_keypoints2d(
        keypoints: Keypoints,
        output_path: str,
        img_arr: Union[None, np.ndarray] = None,
        img_paths: Union[None, List[str]] = None,
        video_path: Union[None, str] = None,
        overwrite: bool = False,
        resolution: Tuple = None,
        return_array: bool = False) -> Union[None, np.ndarray]:
    """Visualize 2d keypoints, powered by mmhuman3d.

    Args:
        keypoints (Keypoints):
            An keypoints2d instance of Keypoints.
        output_path (str):
            Path to the output file. Either a video path
            or a path to an image folder.
        img_arr (Union[None, np.ndarray], optional):
            A single-view image array, in shape
            [n_frame, h, w, c]. Defaults to None.
        img_paths (Union[None, List[str]], optional):
            A list of image paths. Defaults to None.
        video_path (Union[None, str], optional):
            Path to a video file. Defaults to None.
        overwrite (bool, optional):
            Whether replace the file at output_path. Defaults to False.
        return_array (bool, optional):
            Whether to return the visualized image array.
            Defaults to False.

    Raises:
        ValueError: Redundant input.

    Returns:
        Union[None, np.ndarray]:
            If return_array it returns an array of images,
            else return None.
    """
    logger = keypoints.logger
    if not has_mmhuman3d:
        logger.error(import_exception)
        raise ImportError
    # prepare keypoints data
    keypoints_np = keypoints.to_numpy()
    kps2d = keypoints_np.get_keypoints()[..., :2]
    if keypoints_np.get_person_number() == 1:
        kps2d = np.squeeze(kps2d, axis=1)
    kps_convention = keypoints_np.get_convention()
    kps_mask = keypoints_np.get_mask()
    mm_kps_mask = np.sign(np.sum(np.abs(kps_mask), axis=(0, 1)))
    # prepare background data
    input_list = [img_arr, img_paths, video_path]
    input_count = 0
    for input_instance in input_list:
        if input_instance is not None:
            input_count += 1
    if input_count > 1:
        logger.error('Redundant input!\n' + 'Please offer only one among' +
                     ' img_arr, img_paths and video_path.')
        raise ValueError
    vis_arr = visualize_kp2d(
        kp2d=kps2d,
        output_path=output_path,
        image_array=img_arr,
        frame_list=img_paths,
        origin_frames=video_path,
        data_source=kps_convention,
        resolution=resolution,
        mask=mm_kps_mask,
        overwrite=overwrite,
        return_array=return_array)
    return vis_arr



import copy
import os
import time
import torch
import torch.backends.cudnn as cudnn
import torch.nn as nn
from pathlib import Path

from xrmocap.utils.distribute_utils import is_main_process


class AverageMeter(object):
    """Computes and stores the average and current value."""

    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count


def match_name_keywords(n, name_keywords):
    """Match the keys in two dictionaries."""
    for b in name_keywords:
        if b in n:
            return True
    return False


def get_model_name(model, resnet_layer):
    name = f'{model}_{resnet_layer}'
    return name


def get_directory(state: str = 'train',
                  output_dir: str = 'output',
                  cfg_name: str = 'mvp_train',
                  dataset: str = 'dataset',
                  model: str = 'multi_view_pose_transformer',
                  resnet_layer: int = 50):
    """This function generates the log file directory and output file
    directory.

    Args:
        state (str, optional): Train or test mode. Defaults to 'train'.
        output_dir (str, optional): Output directory. Defaults to 'output'.
        cfg_name (str, optional): Name of the configuration file. Defaults
            to 'mvp_train'.
        dataset (str, optional): Name of the dataset. Defaults to 'dataset'.
        model (str, optional): Name of the model. Defaults to
            'multi_view_pose_transformer'.
        resnet_layer (int, optional): Number of layers in the backbone Resnet.
            Defaults to 50.
    """
    this_dir = Path(os.path.dirname(__file__))
    root_output_dir = (this_dir / '..' / '..' / output_dir).resolve()

    if not root_output_dir.exists() and is_main_process():
        root_output_dir.mkdir()

    model = get_model_name(model, resnet_layer)
    cfg_name = os.path.basename(cfg_name).split('.')[0]

    # get final output dir
    final_output_dir = os.path.join(root_output_dir, dataset, model, cfg_name)

    if (not os.path.exists(final_output_dir)):
        os.makedirs(final_output_dir)

    # get logger file under final output dir
    time_str = time.strftime('%Y-%m-%d-%H-%M')
    log_file = f'{cfg_name}_{time_str}_{state}.log'
    log_file = os.path.join(final_output_dir, log_file)

    return log_file, final_output_dir


def save_checkpoint(states,
                    is_best,
                    output_dir,
                    filename='checkpoint.pth.tar'):
    torch.save(states, os.path.join(output_dir, filename))
    if is_best and 'state_dict' in states:
        torch.save(states['state_dict'],
                   os.path.join(output_dir, 'model_best.pth.tar'))


@torch.no_grad()
def accuracy(output, target, topk=(1, )):
    """Computes the precision@k for the specified values of k."""
    if target.numel() == 0:
        return [torch.zeros([], device=output.device)]
    maxk = max(topk)
    batch_size = target.size(0)

    _, pred = output.topk(maxk, 1, True, True)
    pred = pred.t()
    correct = pred.eq(target.view(1, -1).expand_as(pred))

    res = []
    for k in topk:
        correct_k = correct[:k].view(-1).float().sum(0)
        res.append(correct_k.mul_(100.0 / batch_size))
    return res


def get_total_grad_norm(parameters, norm_type=2):
    parameters = list(filter(lambda p: p.grad is not None, parameters))
    norm_type = float(norm_type)
    device = parameters[0].grad.device
    total_norm = torch.norm(
        torch.stack([
            torch.norm(p.grad.detach(), norm_type).to(device)
            for p in parameters
        ]), norm_type)
    return total_norm


def inverse_sigmoid(x, eps=1e-5):
    x = x.clamp(min=0, max=1)
    x1 = x.clamp(min=eps)
    x2 = (1 - x).clamp(min=eps)
    return torch.log(x1 / x2)


def convert_result_to_kps(pred):
    """Convert MVP predict result to Keypoints object."""
    pred = pred[0]
    pred = pred[pred[:, 0, 3] >= 0]
    per_frame_kps3d = pred[:, :, :4]
    n_person = len(per_frame_kps3d)
    return n_person, per_frame_kps3d


def absolute2norm(absolute_coords, grid_size, grid_center):
    """convert abosolute keypoint coordinates to normalized(0-1)
    coordinates."""
    device = absolute_coords.device
    grid_size = grid_size.to(device=device)
    grid_center = grid_center.to(device=device)
    norm_coords = (absolute_coords - grid_center + grid_size / 2.0) / grid_size
    return norm_coords


def norm2absolute(norm_coords, grid_size, grid_center):
    """convert normalized (0-1) keypoint coordinates to abosolute
    coordinates."""
    device = norm_coords.device
    grid_size = grid_size.to(device=device)
    grid_center = grid_center.to(device=device)
    loc = norm_coords * grid_size + grid_center - grid_size / 2.0
    return loc


def get_clones(module, N):
    return nn.ModuleList([copy.deepcopy(module) for i in range(N)])


def set_cudnn(benchmark: bool, deterministic: bool, enable: bool):
    cudnn.benchmark = benchmark
    torch.backends.cudnn.deterministic = deterministic
    torch.backends.cudnn.enabled = enable



# yapf: disable

import logging
import numpy as np
import torch
from typing import Any, Union

from .smplx_data import SMPLXData

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

# yapf: enable


class SMPLXDData(SMPLXData):
    NUM_VERTS = 10475

    def __init__(self,
                 src_dict: dict = None,
                 gender: Union[Literal['female', 'male', 'neutral'],
                               None] = None,
                 fullpose: Union[np.ndarray, torch.Tensor, None] = None,
                 transl: Union[np.ndarray, torch.Tensor, None] = None,
                 betas: Union[np.ndarray, torch.Tensor, None] = None,
                 expression: Union[np.ndarray, torch.Tensor, None] = None,
                 displacement: Union[np.ndarray, torch.Tensor, None] = None,
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """Construct a SMPLXData instance with pre-set values. If any of
        gender, fullpose, transl, betas is provided, it will override the item
        in source_dict.

        Args:
            src_dict (dict, optional):
                A dict with items in HumanData fashion.
                Defaults to None.
            gender (Union[
                    Literal['female', 'male', 'neutral'], None], optional):
                Gender of the body model.
                Should be one among ["female", "male", "neutral"].
                Defaults to None.
            fullpose (Union[np.ndarray, torch.Tensor, None], optional):
                A tensor or ndarray for fullpose, in shape [frame_num, 24, 3].
                Defaults to None, zero-tensor will be created.
            transl (Union[np.ndarray, torch.Tensor, None], optional):
                A tensor or ndarray for translation, in shape [frame_num, 3].
                Defaults to None, zero-tensor will be created.
            betas (Union[np.ndarray, torch.Tensor, None], optional):
                A tensor or ndarray for betas,
                in shape [frame_num, betas_dim].
                Defaults to None,
                zero-tensor in shape [frame_num, 10] will be created.
            expression (Union[np.ndarray, torch.Tensor, None], optional):
                A tensor or ndarray for expression,
                in shape [frame_num, expression_dim].
                Defaults to None,
                zero-tensor in shape [frame_num, 10] will be created.
            displacement (Union[np.ndarray, torch.Tensor, None], optional):
                A tensor or ndarray for displacement,
                in shape [frame_num, NUM_VERTS].
                Defaults to None,
                zero-tensor in shape [frame_num, NUM_VERTS] will be created.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """
        SMPLXData.__init__(
            self,
            src_dict=src_dict,
            gender=gender,
            transl=transl,
            fullpose=fullpose,
            betas=betas,
            expression=expression,
            logger=logger)
        if displacement is None and 'displacement' not in self:
            displacement = np.zeros(
                shape=(self.get_batch_size(), self.__class__.NUM_VERTS))
        if displacement is not None:
            self.set_displacement(displacement)

    def set_displacement(
            self, displacement: Union[np.ndarray, torch.Tensor]) -> None:
        """Set displacement data.

        Args:
            displacement (Union[np.ndarray, torch.Tensor]):
                Displacement parameters in ndarray or tensor,
                in shape [batch_size, NUM_VERTS].

        Raises:
            TypeError: Type of displacement is not correct.
        """
        if isinstance(displacement, torch.Tensor):
            displacement = displacement.detach().cpu().numpy()
        elif not isinstance(displacement, np.ndarray):
            self.logger.error('Type of displacement is not correct.\n' +
                              f'Type: {type(displacement)}.')
            raise TypeError
        if len(displacement.shape) == 1:
            displacement = displacement[np.newaxis, ...]
        displacement_dim = displacement.shape[-1]
        displacement_np = displacement.reshape(-1, displacement_dim)
        dict.__setitem__(self, 'displacement', displacement_np)

    def get_displacement(self) -> np.ndarray:
        """Get displacement.

        Returns:
            ndarray:
                Displacement in shape [batch_size, NUM_VERTS].
        """
        displacement = self.__getitem__('displacement')
        return displacement

    def __setitem__(self, __k: Any, __v: Any) -> None:
        """Set item according to its key.

        Args:
            __k (Any): Key in dict.
            __v (Any): Value in dict.
        """
        if __k == 'displacement':
            self.set_displacement(__v)
        else:
            SMPLXData.__setitem__(self, __k, __v)

    def from_param_dict(self, smplxd_dict: dict) -> None:
        """Load SMPLX+D parameters from smplxd_dict.

        Args:
            smplx_dict (dict):
                A dict of ndarray|Tensor parameters.
                global_orient and body_pose are necessary,
                jaw_pose, leye_pose, reye_pose,
                left_hand_pose, right_hand_pose, transl and
                betas are optional.
                Other keys are ignored.

        Raises:
            KeyError: missing necessary keys.
        """
        necessary_keys = {'global_orient', 'body_pose', 'displacement'}
        if not necessary_keys.issubset(smplxd_dict):
            self.logger.error('Keys are not enough.\n' +
                              f'smplxd_dict\'s keys: {smplxd_dict.keys()}')
            raise KeyError
        SMPLXData.from_param_dict(self, smplxd_dict)
        self.set_displacement(smplxd_dict['displacement'])

    def to_param_dict(self,
                      repeat_betas: bool = True,
                      repeat_expression: bool = True) -> dict:
        """Split fullpose into global_orient, body_pose, jaw_pose, leye_pose,
        reye_pose, left_hand_pose, right_hand_pose, return all the necessary
        parameters in one dict.

        Args:
            repeat_betas (bool, optional):
                Whether to repeat betas when its first dim doesn't match
                batch_size. Defaults to True.
            repeat_expression (bool, optional):
                Whether to repeat expression when its first dim doesn't match
                batch_size. Defaults to True.

        Returns:
            dict:
                A dict of SMPLX data, whose keys are
                betas, global_orient, transl, global_orient, body_pose,
                jaw_pose, leye_pose, reye_pose, left_hand_pose,
                right_hand_pose, expression.
        """
        dict_to_return = SMPLXData.to_param_dict(
            self,
            repeat_betas=repeat_betas,
            repeat_expression=repeat_expression)
        dict_to_return['displacement'] = self.get_displacement()
        return dict_to_return



# yapf: disable
import logging
import numpy as np
import torch
from typing import List, Tuple, Union, overload
from xrprimer.data_structure.camera import FisheyeCameraParameter

from xrmocap.data_structure.body_model import SMPLData
from xrmocap.data_structure.keypoints import Keypoints
from xrmocap.io.image import (
    get_n_frame_from_mview_src, load_clip_from_mview_src,
)
from xrmocap.model.architecture.builder import build_architecture
from xrmocap.model.registrant.builder import SMPLify, build_registrant
from xrmocap.transform.keypoints3d.optim.builder import (
    BaseOptimizer, build_keypoints3d_optimizer,
)
from .mperson_smpl_estimator import MultiPersonSMPLEstimator

# yapf: enable


class MultiViewMultiPersonEnd2EndEstimator(MultiPersonSMPLEstimator):
    """Api for estimating keypoints3d and smpl in a multi-view multi-person
    scene, using end2end learning-based method."""

    def __init__(self,
                 work_dir: str,
                 smplify: Union[dict, SMPLify],
                 kps3d_model: Union[dict, torch.nn.Module],
                 load_batch_size: int = 500,
                 kps3d_optimizers: Union[List[Union[BaseOptimizer, dict]],
                                         None] = None,
                 verbose: bool = True,
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """Initialization of the class.

        Args:
            work_dir (str):
                Path to the folder for running the api. No file in work_dir
                will be modified
                or added by MultiViewSinglePersonSMPLEstimator.
            smplify (Union[dict, SMPLify]):
                A SMPLify instance or its config.
            kps3d_model (Union[dict, torch.nn.Module]):
                An end-to-end mview mperson keypoints3d predicting model.
            kps3d_optimizers (Union[
                    List[Union[BaseOptimizer, dict]], None], optional):
                A list of keypoints3d optimizers or their configs. If given,
                keypoints3d will be
                optimized by the cascaded final optimizers before estimation.
                Defaults to None.
            verbose (bool, optional):
                Whether to print(logger.info) information during estimating.
                Defaults to True.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """
        super().__init__(
            work_dir=work_dir,
            smplify=smplify,
            kps3d_optimizers=kps3d_optimizers,
            verbose=verbose,
            logger=logger)

        self.load_batch_size = load_batch_size
        if isinstance(kps3d_model, dict):
            self.kps3d_model = build_architecture(kps3d_model)
        else:
            self.kps3d_model = kps3d_model

        if isinstance(smplify, dict):
            smplify['logger'] = logger
            self.smplify = build_registrant(smplify)
        else:
            self.smplify = smplify
        if kps3d_optimizers is None:
            self.kps3d_optimizers = None
        else:
            self.kps3d_optimizers = []
            for kps3d_optim in kps3d_optimizers:
                if isinstance(kps3d_optim, dict):
                    kps3d_optim['logger'] = logger
                    kps3d_optim = build_keypoints3d_optimizer(kps3d_optim)
                self.kps3d_optimizers.append(kps3d_optim)

    @overload
    def run(
        self, img_arr: np.ndarray, cam_param: List[FisheyeCameraParameter]
    ) -> Tuple[List[Keypoints], Keypoints, SMPLData]:
        ...

    @overload
    def run(
        self, img_paths: List[List[str]],
        cam_param: List[FisheyeCameraParameter]
    ) -> Tuple[List[Keypoints], Keypoints, SMPLData]:
        ...

    @overload
    def run(
        self, video_path: List[str], cam_param: List[FisheyeCameraParameter]
    ) -> Tuple[List[Keypoints], Keypoints, SMPLData]:
        ...

    def run(
        self,
        cam_param: List[FisheyeCameraParameter],
        img_arr: Union[None, np.ndarray] = None,
        img_paths: Union[None, List[List[str]]] = None,
        video_paths: Union[None, List[str]] = None,
    ) -> Tuple[List[Keypoints], List[SMPLData]]:
        """Run mutli-view multi-person end2end estimator once. run() needs one
        images input among [img_arr, img_paths, video_paths].

        Args:
            cam_param (List[FisheyeCameraParameter]):
                A list of FisheyeCameraParameter instances.
            img_arr (Union[None, np.ndarray], optional):
                A multi-view image array, in shape
                [n_view, n_frame, h, w, c]. Defaults to None.
            img_paths (Union[None, List[List[str]]], optional):
                A nested list of image paths, in shape
                [n_view, n_frame]. Defaults to None.
            video_paths (Union[None, List[str]], optional):
                A list of video paths, each is a view.
                Defaults to None.

        Returns:
            Tuple[List[Keypoints], List[SMPLData]]:
                A list of keypoints3d, a list of SMPLData.
        """
        input_list = [img_arr, img_paths, video_paths]
        input_count = 0
        for input_instance in input_list:
            if input_instance is not None:
                input_count += 1
        if input_count > 1:
            self.logger.error('Redundant input!\n' +
                              'Please offer only one among' +
                              ' img_arr, img_paths and video_paths.')
            raise ValueError
        elif input_count < 1:
            self.logger.error('No image input has been found!\n' +
                              'img_arr, img_paths and video_paths are None.')
            raise ValueError
        n_frame = get_n_frame_from_mview_src(img_arr, img_paths, video_paths,
                                             self.logger)
        for start_idx in range(0, n_frame, self.load_batch_size):
            end_idx = min(n_frame, start_idx + self.load_batch_size)
            mview_batch_arr = load_clip_from_mview_src(
                start_idx=start_idx,
                end_idx=end_idx,
                img_arr=img_arr,
                img_paths=img_paths,
                video_paths=video_paths,
                logger=self.logger)
            n_view, n_frame, h, w, c = mview_batch_arr.shape
            # TODO: infer keypoints3d here
            kps3d_batch = None
        keypoints3d_list = [
            Keypoints(kps=kps3d_batch, mask=np.ones_like(kps3d_batch[..., 0]))
        ]
        smpl_data_list = [
            None,
        ] * len(keypoints3d_list)
        for person_idx, keypoints3d in enumerate(keypoints3d_list):
            keypoints3d = self.optimize_keypoints3d(keypoints3d)
            smpl_data = self.estimate_smpl(keypoints3d)
            keypoints3d_list[person_idx] = keypoints3d
            smpl_data_list[person_idx] = smpl_data
        return keypoints3d_list, smpl_data_list



# copy from mmhuman3d
from abc import ABCMeta, abstractmethod
from mmcv.runner import BaseModule


class BaseArchitecture(BaseModule, metaclass=ABCMeta):
    """Base class for xrmocap architecture."""

    def __init__(self, init_cfg=None):
        super(BaseArchitecture, self).__init__(init_cfg)

    @abstractmethod
    def forward_train(self, **kwargs):
        pass

    @abstractmethod
    def forward_test(self, **kwargs):
        pass

    def _parse_losses(self, losses):
        """Parse the raw outputs (losses) of the network.

        Args:
            losses (dict): Raw output of the network, which usually contain
                losses and other necessary information.
        Returns:
            tuple[Tensor, dict]: (loss, log_vars), loss is the loss tensor \
                which may be a weighted sum of all losses, log_vars contains \
                all the variables to be sent to the logger.
        """
        pass

    def train_step(self, data, optimizer):
        """The iteration step during training.

        This method defines an iteration step during training, except for the
        back propagation and optimizer updating, which are done in an optimizer
        hook. Note that in some complicated cases or models, the whole process
        including back propagation and optimizer updating is also defined in
        this method, such as GAN.

        Args:
            data (dict): The output of dataloader.
            optimizer (:obj:`torch.optim.Optimizer` | dict): The optimizer of
                runner is passed to ``train_step()``. This argument is unused
                and reserved.

        Returns:
            dict: It should contain at least 3 keys: ``loss``, ``log_vars``,
                ``num_samples``.

                - ``loss`` is a tensor for back propagation, which can be a
                  weighted sum of multiple losses.
                - ``log_vars`` contains all the variables to be sent to the
                  logger.
                - ``num_samples`` indicates the batch size (when the model is
                  DDP, it means the batch size on each GPU), which is used for
                  averaging the logs.
        """
        pass

    def val_step(self, data, optimizer=None):
        """The iteration step during validation.

        This method shares the same signature as :func:`train_step`, but used
        during val epochs. Note that the evaluation after training epochs is
        not implemented with this method, but an evaluation hook.
        """
        pass

    def forward(self, **kwargs):
        if self.training:
            return self.forward_train(**kwargs)
        else:
            return self.forward_test(**kwargs)



# yapf: disable
import cv2
import logging
import numpy as np
from typing import List, Union
from xrprimer.utils.ffmpeg_utils import VideoWriter
from xrprimer.utils.log_utils import get_logger

# yapf: enable


def mview_array_to_video(
        mview_img_arr: Union[np.ndarray, List[np.ndarray]],
        output_path: str,
        logger: Union[None, str, logging.Logger] = None) -> None:
    """Concat multi-view video array together, align them into grid, and write
    it to a single video file.

    Args:
        mview_img_arr (Union[np.ndarray, List[np.ndarray]]):
            Am array or a nested list of multi-view
            image array, in shape [n_view, n_frame, h, w, n_ch].
        output_path (str):
            Path to the output video file.
        logger (Union[None, str, logging.Logger], optional):
            Logger for logging. If None, root logger will be selected.
            Defaults to None.
    """
    logger = get_logger(logger)
    n_view = len(mview_img_arr)
    sview_img = mview_img_arr[0]
    if len(sview_img.shape) != 4:
        logger.error('Shape of mview_img_arr should be' +
                     ' [n_view, n_frame, h, w, n_ch].\n' +
                     f'mview_img_arr.shape: {n_view, *sview_img.shape}.')
        raise ValueError
    # how many video grid along x and y
    n_video_y = np.sqrt(n_view)
    if n_video_y - int(n_video_y) > 0.001:
        n_video_y = int(n_video_y)
        n_video_x = n_video_y + 1
        if n_video_x * n_video_y < n_view:
            n_video_y += 1
    else:
        n_video_y = int(n_video_y)
        n_video_x = n_video_y
    # use the max n_frame and max resolution
    # as template
    n_frames = 0
    grid_h = 0
    grid_w = 0
    n_ch = 0
    for _, sview_img_arr in enumerate(mview_img_arr):
        n_frames = max(len(sview_img_arr), n_frames)
        h, w, ch = sview_img_arr[0].shape
        grid_h = max(h, grid_h)
        grid_w = max(w, grid_w)
        n_ch = max(ch, n_ch)
    aio_h = grid_h * n_video_y
    aio_w = grid_w * n_video_x
    v_writer = VideoWriter(
        output_path=output_path,
        resolution=[aio_h, aio_w],
        n_frames=n_frames,
        logger=logger)
    for frame_idx in range(n_frames):
        canvas = np.zeros(shape=(aio_h, aio_w, n_ch), dtype=np.uint8)
        for view_idx in range(n_view):
            x_start = (view_idx % n_video_x) * grid_w
            y_start = int(view_idx / n_video_x) * grid_h
            if frame_idx < len(mview_img_arr[view_idx]):
                content = cv2.resize(mview_img_arr[view_idx][frame_idx],
                                     [grid_w, grid_h])
                canvas[y_start:y_start + grid_h,
                       x_start:x_start + grid_w, :] = content
        v_writer.write(canvas)
    v_writer.close()



import cv2
import logging
import numpy as np
from typing import List, Union
from xrprimer.utils.ffmpeg_utils import VideoInfoReader, video_to_array


def load_multiview_images(
        img_paths: Union[None, List[List[str]]]) -> np.ndarray:
    """Load multi-view images to an ndarray.

    Args:
        img_paths (Union[None, List[List[str]]]):
            A nested list of image paths, in shape
            [view_n, frame_n].

    Returns:
        np.ndarray:
            Multi-view image array, in shape
            [view_n, frame_n, h, w, c].
    """
    # multi-view list
    mview_list = []
    for view_list in img_paths:
        # single-view list
        sv_list = []
        for img_path in view_list:
            frame_np = cv2.imread(filename=img_path)
            sv_list.append(frame_np)
        mview_list.append(sv_list)
    mview_array = np.asarray(mview_list)
    return mview_array


def get_n_frame_from_mview_src(
        img_arr: Union[None, np.ndarray] = None,
        img_paths: Union[None, List[List[str]]] = None,
        video_paths: Union[None, List[str]] = None,
        logger: Union[None, str, logging.Logger] = None) -> int:
    """Get number of frames from multi-view image source. It needs one images
    input among [img_arr, img_paths, video_paths].

    Args:
        img_arr (Union[None, np.ndarray], optional):
            A multi-view image array, in shape
            [n_view, n_frame, h, w, c]. Defaults to None.
        img_paths (Union[None, List[List[str]]], optional):
            A nested list of image paths, in shape
            [n_view, n_frame]. Defaults to None.
        video_paths (Union[None, List[str]], optional):
            A list of video paths, each is a view.
            Defaults to None.
        logger (Union[None, str, logging.Logger], optional):
            Logger for logging. If None, root logger will be selected.
            Defaults to None.

    Returns:
        int: Number of frames.
    """
    n_frame = None
    if img_arr is not None:
        n_frame = img_arr.shape[1]
    elif img_paths is not None:
        n_frame = len(img_paths[0])
    elif video_paths is not None:
        reader = VideoInfoReader(video_paths[0], logger=logger)
        n_frame = int(reader['nb_frames'])
    return n_frame


def load_clip_from_mview_src(
        start_idx: int,
        end_idx: int,
        img_arr: Union[None, np.ndarray] = None,
        img_paths: Union[None, List[List[str]]] = None,
        video_paths: Union[None, List[str]] = None,
        logger: Union[None, str, logging.Logger] = None) -> np.ndarray:
    """Get image array of a clip from multi-view image source. It needs one
    images input among [img_arr, img_paths, video_paths].

    Args:
        img_arr (Union[None, np.ndarray], optional):
            A multi-view image array, in shape
            [n_view, n_frame, h, w, c]. Defaults to None.
        img_paths (Union[None, List[List[str]]], optional):
            A nested list of image paths, in shape
            [n_view, n_frame]. Defaults to None.
        video_paths (Union[None, List[str]], optional):
            A list of video paths, each is a view.
            Defaults to None.
        logger (Union[None, str, logging.Logger], optional):
            Logger for logging. If None, root logger will be selected.
            Defaults to None.

    Returns:
        np.ndarray: Image array of the selected range.
    """
    ret_img_arr = None
    if img_arr is not None:
        ret_img_arr = img_arr[:, start_idx:end_idx, ...]
    elif img_paths is not None:
        ret_img_list = []
        for view_idx in range(len(img_paths)):
            view_img_list = []
            for frame_idx in range(start_idx, end_idx):
                img = cv2.imread(img_paths[view_idx][frame_idx])
                view_img_list.append(img)
            ret_img_list.append(view_img_list)
        ret_img_arr = np.asarray(ret_img_list)
    elif video_paths is not None:
        ret_img_list = []
        for view_idx in range(len(video_paths)):
            img_arr = video_to_array(
                video_paths[view_idx],
                start=start_idx,
                end=end_idx,
                logger=logger)
            ret_img_list.append(img_arr)
        ret_img_arr = np.asarray(ret_img_list)
    return ret_img_arr



# yapf: disable
import json
import logging
import numpy as np
from scipy.spatial.transform import Rotation as scipy_Rotation
from typing import List, Tuple, Union
from xrprimer.data_structure.camera import FisheyeCameraParameter
from xrprimer.transform.camera.extrinsic import rotate_camera
from xrprimer.utils.log_utils import get_logger

from xrmocap.data_structure.smc_reader import SMCReader

# yapf: enable

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


def load_camera_parameters_from_zoemotion_dir(
        camera_parameter_path: str,
        enable_camera_id: Union[list, None] = None) -> Tuple[list, list]:
    """Load camera parameter and get an RGB FisheyeCameraParameter.

    Args:
        camera_parameter_path (str): path to the camera parameter
        enable_camera_id (Union[list, None], optional): camera ID(str).
        Defaults to None.

    Returns:
        cam_param_list (list): FisheyeCameraParameter
        enable_camera_list (list): enable camera list e.g.['0', '1']
    """
    enable_camera_list = []
    camera_param_dict = json.load(open(camera_parameter_path))
    if enable_camera_id is None:
        enable_camera_list = [str(x) for x in range(len(camera_param_dict))]
    else:
        enable_camera_list = enable_camera_id.split('_')

    cam_param_list = []
    for camera_id in enable_camera_list:
        dist_coeff_k = []
        dist_coeff_p = []
        cam_param = FisheyeCameraParameter()
        cam_param.name = camera_id
        dist_coeff_k = [
            camera_param_dict[camera_id]['distCoeff'][0],
            camera_param_dict[camera_id]['distCoeff'][1],
            camera_param_dict[camera_id]['distCoeff'][4], 0, 0, 0
        ]
        dist_coeff_p = [
            camera_param_dict[camera_id]['distCoeff'][2],
            camera_param_dict[camera_id]['distCoeff'][3]
        ]
        cam_param.set_dist_coeff(dist_coeff_k, dist_coeff_p)
        cam_param.set_KRT(
            K=np.array(camera_param_dict[camera_id]['K']).reshape(3, 3),
            R=np.array(camera_param_dict[camera_id]['R']).reshape(3, 3),
            T=camera_param_dict[camera_id]['T'],
            world2cam=False)
        cam_param.set_resolution(
            height=camera_param_dict[camera_id]['imgSize'][1],
            width=camera_param_dict[camera_id]['imgSize'][0])
        cam_param_list.append(cam_param)

    return cam_param_list, enable_camera_list


def get_color_camera_parameter_from_smc(
        smc_reader: SMCReader,
        camera_type: Literal['kinect', 'iphone'],
        camera_id: int,
        align_floor: bool = True,
        logger: Union[None, str,
                      logging.Logger] = None) -> FisheyeCameraParameter:
    """Get an RGB FisheyeCameraParameter from an smc reader.

    Args:
        smc_reader (SMCReader):
            An SmcReader instance containing kinect
            and iphone camera parameters.
        camera_type (Literal['kinect', 'iphone']):
            Which type of camera to get.
        camera_id (int):
            ID of the selected camera.
        align_floor (bool):
            Whether to rotate camera extrinsics, makes the xOz
            plane parallel to floor calibrated by Kinect_0's
            depth camera.
            Defaults to True.
        logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.

    Raises:
        NotImplementedError: iphone has not been supported yet.
        KeyError: camera_type is neither kinect nor iphone.

    Returns:
        FisheyeCameraParameter
    """
    logger = get_logger(logger)
    cam_param = FisheyeCameraParameter(name=f'{camera_type}_{camera_id:02d}')
    if camera_type == 'kinect':
        extrinsics_dict = \
            smc_reader.get_kinect_color_extrinsics(
                camera_id, homogeneous=False
            )
        extrinsics_r_np = extrinsics_dict['R'].reshape(3, 3)
        extrinsics_t_np = extrinsics_dict['T'].reshape(3)
        intrinsics_np = \
            smc_reader.get_kinect_color_intrinsics(
                camera_id
            )
        resolution = \
            smc_reader.get_kinect_color_resolution(
                camera_id
            )
        cam_param.set_KRT(
            K=intrinsics_np.tolist(),
            R=extrinsics_r_np.tolist(),
            T=extrinsics_t_np.tolist(),
            world2cam=False)
        cam_param.set_resolution(
            height=int(resolution[1]), width=int(resolution[0]))
    elif camera_type == 'iphone':
        raise NotImplementedError('iphone has not been supported yet.')
    else:
        logger.error('Choose camera_type from [\'kinect\', \'iphone\'].')
        raise KeyError('Wrong camera_type.')
    if align_floor:
        floor_dict = smc_reader.get_depth_floor(0)
        calib_space_normal = np.array(floor_dict['normal']).reshape((3))
        opencv_space_normal = np.asarray([0, -1, 0])
        rotation = scipy_Rotation.align_vectors(
            np.expand_dims(opencv_space_normal, 0),
            np.expand_dims(calib_space_normal, 0))[0]
        cam_rot_mat = rotation.as_matrix()
        cam_param = rotate_camera(cam_param, cam_rot_mat)
    return cam_param


def get_all_color_kinect_parameter_from_smc(
    smc_reader: SMCReader,
    align_floor: bool = True,
    logger: Union[None, str, logging.Logger] = None
) -> List[FisheyeCameraParameter]:
    """Get an RGB FisheyeCameraParameter from an smc reader.

    Args:
        smc_reader (SMCReader):
            An SmcReader instance containing kinect
            and iphone camera parameters.
        align_floor (bool):
            Whether to rotate camera extrinsics, makes the xOz
            plane parallel to floor calibrated by Kinect_0's
            depth camera.
            Defaults to True.
        logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.

    Returns:
        List[FisheyeCameraParameter]
    """
    logger = get_logger(logger)
    ret_list = []
    for kinect_index in range(smc_reader.num_kinects):
        cam_param = get_color_camera_parameter_from_smc(
            smc_reader=smc_reader,
            camera_type='kinect',
            camera_id=kinect_index,
            align_floor=False,
            logger=logger)
        ret_list.append(cam_param)
    if align_floor:
        floor_dict = smc_reader.get_depth_floor(0)
        calib_space_normal = np.array(floor_dict['normal']).reshape((3))
        opencv_space_normal = np.asarray([0, -1, 0])
        rotation = scipy_Rotation.align_vectors(
            np.expand_dims(opencv_space_normal, 0),
            np.expand_dims(calib_space_normal, 0))[0]
        cam_rot_mat = rotation.as_matrix()
        for kinect_index, cam_param in enumerate(ret_list):
            ret_list[kinect_index] = rotate_camera(cam_param, cam_rot_mat)
    return ret_list



import torch

from xrmocap.transform.convention.keypoints_convention import \
    get_keypoint_idxs_by_part  # noqa:E501
from .smplify import SMPLify


class SMPLifyX(SMPLify):
    """Re-implementation of SMPLify-X with extended features."""
    OPTIM_PARAM = SMPLify.OPTIM_PARAM + [
        'left_hand_pose', 'right_hand_pose', 'expression', 'jaw_pose',
        'leye_pose', 'reye_pose'
    ]

    def __set_keypoint_indexes__(self) -> None:
        """Set keypoint indexes to 1) body parts to be assigned different
        weights 2) be ignored for keypoint loss computation.

        Returns:
            None
        """
        SMPLify.__set_keypoint_indexes__(self)
        convention = self.body_model.keypoint_convention
        # head keypoints include all facial landmarks
        self.face_keypoint_idxs = get_keypoint_idxs_by_part(
            'head', convention=convention)

        left_hand_keypoint_idxs = get_keypoint_idxs_by_part(
            'left_hand', convention=convention)
        right_hand_keypoint_idxs = get_keypoint_idxs_by_part(
            'right_hand', convention=convention)
        self.hand_keypoint_idxs = [
            *left_hand_keypoint_idxs, *right_hand_keypoint_idxs
        ]

        self.body_keypoint_idxs = get_keypoint_idxs_by_part(
            'body', convention=convention)

    def get_keypoint_weight(self,
                            use_shoulder_hip_only: bool = False,
                            body_weight: float = 1.0,
                            hand_weight: float = 1.0,
                            face_weight: float = 1.0) -> torch.Tensor:
        """Get per keypoint weight.

        Args:
            use_shoulder_hip_only (bool, optional):
                Whether to use only shoulder and hip
                keypoints for loss computation. This is useful in the
                warming-up stage to find a reasonably good initialization.
                Defaults to False.
            body_weight (float, optional):
                Weight of body keypoints. Body part segmentation
                definition is included in the HumanData convention.
                Defaults to 1.0.

        Returns:
            torch.Tensor: Per keypoint weight tensor of shape (K).
        """
        n_keypoints = self.body_model.get_joint_number()

        if use_shoulder_hip_only:
            weight = torch.zeros([n_keypoints]).to(self.device)
            weight[self.shoulder_hip_keypoint_idxs] = 1.0
        else:
            weight = torch.ones([n_keypoints]).to(self.device)

            weight[self.body_keypoint_idxs] = \
                weight[self.body_keypoint_idxs] * body_weight
            weight[self.hand_keypoint_idxs] = \
                weight[self.hand_keypoint_idxs] * hand_weight
            weight[self.face_keypoint_idxs] = \
                weight[self.face_keypoint_idxs] * face_weight

        if hasattr(self, 'ignore_keypoint_idxs'):
            weight[self.ignore_keypoint_idxs] = 0.0

        return weight



import logging
import torch
from typing import Any, Union
from xrprimer.utils.log_utils import get_logger


class BaseImageTransform(torch.nn.Module):

    def __init__(self,
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """Base class for image transform.

        Args:
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """
        super().__init__()
        self.logger = get_logger(logger)

    def forward(self, input: Any) -> Any:
        """Forward function of ImageTransform.

        Args:
            input (Any)

        Returns:
            Any
        """
        raise NotImplementedError



# yapf: disable

import logging
import numpy as np
import os
import random
import time
import torch
import torch.distributed as dist
import torch.optim as optim
import torch.utils.data
import torch.utils.data.distributed
from mmcv.runner import get_dist_info, load_checkpoint
from prettytable import PrettyTable
from torch.utils.data import DistributedSampler
from typing import Union
from xrprimer.utils.log_utils import get_logger

from xrmocap.core.evaluation.builder import build_evaluation
from xrmocap.data.dataset.builder import build_dataset
from xrmocap.data_structure.keypoints import Keypoints
from xrmocap.model.architecture.builder import build_architecture
from xrmocap.utils.distribute_utils import (
    collect_results, get_rank, is_main_process, time_synchronized,
)
from xrmocap.utils.mvp_utils import (
    AverageMeter, convert_result_to_kps, get_total_grad_norm,
    match_name_keywords, norm2absolute, save_checkpoint, set_cudnn,
)

# yapf: enable


class MVPTrainer():

    def __init__(self,
                 distributed: bool,
                 model_path: Union[None, str],
                 gpu_idx: int,
                 train_dataset: str,
                 test_dataset: str,
                 cudnn_setup: dict,
                 dataset_setup: dict,
                 mvp_setup: dict,
                 pretrained_backbone: Union[None, str] = None,
                 finetune_model: Union[None, str] = None,
                 resume: bool = False,
                 final_output_dir: Union[None, str] = './output',
                 lr_decay_epoch: list = [30],
                 test_model_file: Union[None, str] = None,
                 end_epoch: int = 30,
                 optimizer: Union[None, str] = None,
                 weight_decay: float = 0.0,
                 lr: float = 0.2,
                 logger: Union[None, str, logging.Logger] = None,
                 device: str = 'cuda',
                 seed: int = 42,
                 workers: int = 4,
                 train_batch_size: int = 1,
                 test_batch_size: int = 1,
                 lr_linear_proj_mult: float = 0.1,
                 model_root: str = './weight',
                 inference_conf_thr: list = [0.0],
                 print_freq: int = 100,
                 clip_max_norm: float = 0.1,
                 optimize_backbone: bool = False) -> None:
        """Create a trainer for training the Multi-view Pose Transformer(MVP).

        Args:
            distributed (bool):
                If distributed training is used.
            model_path (Union[None, str]):
                Path to the model weight,for evaluation.
            gpu_idx (int):
                Index of current GPU when using distributed training.
            train_dataset (str):
                Name of the train dataset.
            test_dataset (str):
                Name of the test dataset.
            lr (float, optional):
                Learning rate. Defaults to 0.2.
            weight_decay (float, optional):
                Weight decay. Defaults to 0.0.
            optimizer (Union[None, str], optional):
                Type of optimizer. Defaults to None.
            end_epoch (int, optional):
                End epoch of trainig. Defaults to 30.
            pretrained_backbone (Union[None, str], optional):
                Path to pretrained backbone weights
                if using pretrained model. Defaults to None.
            finetune_model (Union[None, str], optional):
                Path to a pretrained model weights to be finetuned.
                Defaults to None.
            resume (bool, optional):
                If auto resume from checkpoints is used.
                Defaults to False.
            final_output_dir (Union[None, str], optional):
                Path to output folder. Defaults to './output'.
            lr_decay_epoch (list, optional):
                Lr decay milestones. Defaults to [30].
            test_model_file (Union[None, str], optional):
                Path to test model weight. Defaults to None.
            cudnn_setup (dict):
                Dict of parameters to setup cudnn.
            dataset_setup (dict):
                Dict of parameters to setup the dataset.
            mvp_setup (dict):
                Dict of parameters to setup the MVP architecture.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
            device (str, optional):
                Training device, 'cuda' for GPU usage.
                Defaults to 'cuda'.
            seed (int, optional):
                Fix random seed between experiments if needed.
                Defaults to 42.
            workers (int, optional):
                Number of workers.. Defaults to 4.
            train_batch_size (int, optional):
                Batch size in training. Defaults to 1.
            test_batch_size (int, optional):
                Batch size in testing. Defaults to 1.
            lr_linear_proj_mult (float, optional):
                LR factor for linear projection related weights.
                Defaults to 0.1.
            model_root (str, optional):
                Root folder for pretrained weights. Defaults to './weight'.
            inference_conf_thr (list, optional):
                List of confidence threshold to filter non-human
                keypoints. Defaults to [0.0].
            print_freq (int, optional):
                Printing frequency during training. Defaults to 100.
            clip_max_norm (float, optional):
                Gradient clipping. Defaults to 0.1.
            optimize_backbone (bool, optional):
                Set it to be True to train the whole model jointly.
                Defaults to False.
        """

        self.logger = get_logger(logger)
        self.device = device
        self.seed = seed
        self.distributed = distributed
        self.model_path = model_path
        self.gpu_idx = gpu_idx

        self.train_dataset = train_dataset
        self.test_dataset = test_dataset
        self.train_batch_size = train_batch_size
        self.test_batch_size = test_batch_size
        self.lr = lr
        self.lr_linear_proj_mult = lr_linear_proj_mult
        self.weight_decay = weight_decay
        self.optimizer = optimizer
        self.end_epoch = end_epoch
        self.pretrained_backbone = pretrained_backbone
        self.model_root = model_root
        self.finetune_model = finetune_model
        self.resume = resume
        self.lr_decay_epoch = lr_decay_epoch
        self.inference_conf_thr = inference_conf_thr
        self.test_model_file = test_model_file
        self.clip_max_norm = clip_max_norm
        self.print_freq = print_freq
        self.workers = workers
        self.final_output_dir = final_output_dir
        self.optimize_backbone = optimize_backbone

        self.cudnn_setup = cudnn_setup
        self.dataset_setup = dataset_setup
        self.mvp_setup = mvp_setup

        seed = self.seed + get_rank()
        torch.manual_seed(seed)
        np.random.seed(seed)
        random.seed(seed)

    def get_optimizer(self, model_without_ddp, weight_decay, optim_type):
        lr = self.lr
        if model_without_ddp.backbone is not None:
            for params in model_without_ddp.backbone.parameters():
                # Set it to be True to train the whole model jointly
                # Default to false to avoid OOM
                params.requires_grad = self.optimize_backbone

        lr_linear_proj_mult = self.lr_linear_proj_mult
        lr_linear_proj_names = ['reference_points', 'sampling_offsets']
        param_dicts = [{
            'params': [
                p for n, p in model_without_ddp.named_parameters()
                if not match_name_keywords(n, lr_linear_proj_names)
                and p.requires_grad
            ],
            'lr':
            lr,
        }, {
            'params': [
                p for n, p in model_without_ddp.named_parameters()
                if match_name_keywords(n, lr_linear_proj_names)
                and p.requires_grad
            ],
            'lr':
            lr * lr_linear_proj_mult,
        }]

        if optim_type == 'adam':
            optimizer = optim.Adam(param_dicts, lr=lr)
        elif optim_type == 'adamw':
            optimizer = optim.AdamW(param_dicts, lr=lr, weight_decay=1e-4)

        return optimizer

    def train(self):

        if is_main_process():
            self.logger.info('Loading data ..')

        train_dataset_cfg = dict(type='MVPDataset', logger=self.logger)
        train_dataset_cfg.update(self.dataset_setup.train_dataset_setup)
        train_dataset_cfg.update(self.dataset_setup.base_dataset_setup)
        train_dataset = build_dataset(train_dataset_cfg)

        test_dataset_cfg = dict(type='MVPDataset', logger=self.logger)
        test_dataset_cfg.update(self.dataset_setup.test_dataset_setup)
        test_dataset_cfg.update(self.dataset_setup.base_dataset_setup)
        test_dataset = build_dataset(test_dataset_cfg)

        n_views = train_dataset.n_views

        if self.distributed:
            rank, world_size = get_dist_info()
            sampler_train = DistributedSampler(train_dataset)
            sampler_val = DistributedSampler(
                test_dataset, world_size, rank, shuffle=False)
        else:
            sampler_train = torch.utils.data.RandomSampler(train_dataset)
            sampler_val = torch.utils.data.SequentialSampler(test_dataset)

        batch_sampler_train = torch.utils.data.BatchSampler(
            sampler_train, self.train_batch_size, drop_last=True)

        train_loader = torch.utils.data.DataLoader(
            train_dataset,
            batch_sampler=batch_sampler_train,
            num_workers=self.workers,
            pin_memory=True)

        test_loader = torch.utils.data.DataLoader(
            test_dataset,
            batch_size=self.test_batch_size,
            sampler=sampler_val,
            pin_memory=True,
            num_workers=self.workers)

        set_cudnn(self.cudnn_setup.benchmark, self.cudnn_setup.deterministic,
                  self.cudnn_setup.enable)

        if is_main_process():
            self.logger.info('Constructing models ..')

        mvp_cfg = dict(
            type='MviewPoseTransformer', is_train=True, logger=self.logger)
        mvp_cfg.update(self.mvp_setup)
        model = build_architecture(mvp_cfg)

        model.to(self.device)
        model.criterion.to(self.device)

        model_without_ddp = model
        if self.distributed:
            if is_main_process():
                self.logger.info('Distributed ..')
            model = torch.nn.parallel.\
                DistributedDataParallel(model, device_ids=[self.gpu_idx],
                                        find_unused_parameters=True)
            model_without_ddp = model.module

        optimizer = self.get_optimizer(model_without_ddp, self.weight_decay,
                                       self.optimizer)

        end_epoch = self.end_epoch

        if self.pretrained_backbone:
            # Load pretrained poseresnet weight for panoptic only
            checkpoint_file = os.path.join(self.model_root,
                                           self.pretrained_backbone)
            load_checkpoint(
                model_without_ddp.backbone,
                checkpoint_file,
                map_location='cpu',
                logger=self.logger)
        if self.finetune_model is not None:
            # Load the checkpoint with state_dict only
            checkpoint_file = os.path.join(self.model_root,
                                           self.finetune_model)
            checkpoint = load_checkpoint(
                model_without_ddp,
                checkpoint_file,
                map_location='cpu',
                logger=self.logger)
            start_epoch = 0
            best_precision = checkpoint['precision'] \
                if 'precision' in checkpoint else 0
        if self.resume:
            # Load the checkpoint with full dict keys
            checkpoint_file = os.path.join(self.final_output_dir,
                                           'checkpoint.pth.tar')
            checkpoint = load_checkpoint(
                model_without_ddp,
                checkpoint_file,
                map_location='cpu',
                logger=self.logger)
            start_epoch = checkpoint['epoch']
            best_precision = checkpoint['precision'] \
                if 'precision' in checkpoint else 0
            optimizer.load_state_dict(checkpoint['optimizer'])

        else:
            start_epoch, checkpoint, best_precision = 0, None, 0

        # list for step decay
        if isinstance(self.lr_decay_epoch, list):
            lr_scheduler = torch.optim.lr_scheduler.MultiStepLR(
                optimizer, milestones=self.lr_decay_epoch, gamma=0.1)
            if checkpoint is not None and 'lr_scheduler' in checkpoint:
                lr_scheduler.load_state_dict(checkpoint['lr_scheduler'])
        # int for cosine decay
        else:
            lr_scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(
                optimizer, self.lr_decay_epoch, eta_min=1e-5)
            if checkpoint is not None and 'lr_scheduler' in checkpoint:
                lr_scheduler.load_state_dict(checkpoint['lr_scheduler'])

        n_parameters = sum(p.numel() for p in model.parameters()
                           if p.requires_grad)
        if is_main_process():
            self.logger.info(f'Number of params: {n_parameters}')

        for epoch in range(start_epoch, end_epoch):
            current_lr = optimizer.param_groups[0]['lr']
            if is_main_process():
                self.logger.info(f'Epoch: {epoch}, current lr: {current_lr}')
            train_3d(
                model,
                optimizer,
                train_loader,
                epoch,
                self.logger,
                self.final_output_dir,
                self.clip_max_norm,
                self.print_freq,
                n_views=n_views,
                device=self.device)

            lr_scheduler.step()

            for thr in self.inference_conf_thr:
                preds_single, _ = validate_3d(
                    model,
                    test_loader,
                    self.logger,
                    self.final_output_dir,
                    thr,
                    self.print_freq,
                    n_views=n_views,
                    is_train=True)
                preds = collect_results(preds_single, len(test_dataset))

                if is_main_process():
                    precision = None

                    if 'panoptic' in self.test_dataset:
                        tb = PrettyTable()
                        mpjpe_threshold = np.arange(25, 155, 25)

                        eval_cfg = dict(
                            type='MVPEvaluation', dataset=test_loader.dataset)
                        evaluator = build_evaluation(eval_cfg)
                        aps, recs, mpjpe, recall500 = \
                            evaluator.evaluate_map(preds)

                        tb.field_names = ['Threshold/mm'] + \
                            [f'{i}' for i in mpjpe_threshold]
                        tb.add_row(['AP'] + [f'{ap * 100:.2f}' for ap in aps])
                        tb.add_row(['Recall'] +
                                   [f'{re * 100:.2f}' for re in recs])
                        tb.add_row(['recall@500mm'] +
                                   [f'{recall500 * 100:.2f}' for re in recs])
                        self.logger.info('\n' + tb.get_string())
                        self.logger.info(f'MPJPE: {mpjpe:.2f}mm')
                        precision = np.mean(aps[0])

                    elif 'campus' in self.test_dataset \
                            or 'shelf' in self.test_dataset:
                        eval_cfg = dict(
                            type='MVPEvaluation', dataset=test_loader.dataset)
                        evaluator = build_evaluation(eval_cfg)
                        actor_pcp, avg_pcp, recall500 = evaluator.evaluate_pcp(
                            preds, recall_threshold=500, alpha=0.5)
                        tb = PrettyTable()
                        tb.field_names = [
                            'Metric', 'Actor 1', 'Actor 2', 'Actor 3',
                            'Average'
                        ]
                        tb.add_row([
                            'PCP', f'{actor_pcp[0] * 100:.2f}',
                            f'{actor_pcp[1] * 100:.2f}',
                            f'{actor_pcp[2] * 100:.2f}', f'{avg_pcp * 100:.2f}'
                        ])
                        self.logger.info('\n' + tb.get_string())
                        self.logger.info(f'Recall@500mm: {recall500:.4f}')

                        precision = np.mean(avg_pcp)

                    if precision > best_precision:
                        best_precision = precision
                        best_model = True
                    else:
                        best_model = False
                    if isinstance(self.lr_decay_epoch, list):
                        self.logger.info(
                            f'saving checkpoint to {self.final_output_dir} '
                            f'(Best: {best_model})')
                        save_checkpoint(
                            {
                                'epoch': epoch + 1,
                                'state_dict': model.module.state_dict(),
                                'lr_scheduler': lr_scheduler.state_dict(),
                                'precision': best_precision,
                                'optimizer': optimizer.state_dict(),
                            }, best_model, self.final_output_dir)
                    else:
                        self.logger.info(
                            f'saving checkpoint to {self.final_output_dir} '
                            f'(Best: {best_model})')
                        save_checkpoint(
                            {
                                'epoch': epoch + 1,
                                'state_dict': model.module.state_dict(),
                                'precision': best_precision,
                                'optimizer': optimizer.state_dict(),
                            }, best_model, self.final_output_dir)
                dist.barrier()

        if is_main_process():
            final_model_state_file = os.path.join(self.final_output_dir,
                                                  'final_state.pth.tar')
            self.logger.info(
                f'saving final model state to {final_model_state_file}')
            torch.save(model.module.state_dict(), final_model_state_file)

    def eval(self):

        if is_main_process():
            self.logger.info('Loading data ..')

        test_dataset_cfg = dict(type='MVPDataset', logger=self.logger)
        test_dataset_cfg.update(self.dataset_setup.test_dataset_setup)
        test_dataset_cfg.update(self.dataset_setup.base_dataset_setup)
        test_dataset = build_dataset(test_dataset_cfg)

        if self.distributed:
            rank, world_size = get_dist_info()
            sampler_val = DistributedSampler(
                test_dataset, world_size, rank, shuffle=False)
        else:
            sampler_val = torch.utils.data.SequentialSampler(test_dataset)

        test_loader = torch.utils.data.DataLoader(
            test_dataset,
            batch_size=self.test_batch_size,
            sampler=sampler_val,
            pin_memory=True,
            num_workers=self.workers)

        n_views = test_dataset.n_views

        set_cudnn(self.cudnn_setup.benchmark, self.cudnn_setup.deterministic,
                  self.cudnn_setup.enable)

        if is_main_process():
            self.logger.info('Constructing models ..')

        mvp_cfg = dict(
            type='MviewPoseTransformer', is_train=False, logger=self.logger)
        mvp_cfg.update(self.mvp_setup)
        model = build_architecture(mvp_cfg)

        model.to(self.device)
        model.criterion.to(self.device)

        if self.distributed:
            model = torch.nn.parallel.DistributedDataParallel(
                model, device_ids=[self.gpu_idx], find_unused_parameters=True)

        if self.model_path is not None:
            if is_main_process():
                self.logger.info(f'Load saved models state {self.model_path}')

            load_checkpoint(
                model.module,
                self.model_path,
                logger=self.logger,
                map_location='cpu')

        elif os.path.isfile(
                os.path.join(self.final_output_dir, self.test_model_file)):
            test_model_file = \
                os.path.join(self.final_output_dir,
                             self.test_model_file)
            if is_main_process():
                self.logger.info(
                    f'Load default best models state {test_model_file}')
            model.module.load_state_dict(torch.load(test_model_file))
        else:
            raise ValueError('Check the model file for testing!')

        for thr in self.inference_conf_thr:
            preds_single, kps3d = validate_3d(
                model,
                test_loader,
                self.logger,
                self.final_output_dir,
                thr,
                self.print_freq,
                n_views=n_views)
            preds = collect_results(preds_single, len(test_dataset))

            if is_main_process():
                if 'panoptic' in self.test_dataset:
                    tb = PrettyTable()
                    mpjpe_threshold = np.arange(25, 155, 25)

                    eval_cfg = dict(
                        type='MVPEvaluation', dataset=test_loader.dataset)
                    evaluator = build_evaluation(eval_cfg)
                    aps, recs, mpjpe, recall500 = \
                        evaluator.evaluate_map(preds)

                    tb.field_names = ['Threshold/mm'] + \
                                     [f'{i}' for i in mpjpe_threshold]
                    tb.add_row(['AP'] + [f'{ap * 100:.2f}' for ap in aps])
                    tb.add_row(['Recall'] + [f'{re * 100:.2f}' for re in recs])
                    tb.add_row(['recall@500mm'] +
                               [f'{recall500 * 100:.2f}' for re in recs])
                    self.logger.info('\n' + tb.get_string())
                    self.logger.info(f'MPJPE: {mpjpe:.2f}mm')

                elif 'campus' in self.test_dataset \
                        or 'shelf' in self.test_dataset:

                    eval_cfg = dict(
                        type='MVPEvaluation', dataset=test_loader.dataset)
                    evaluator = build_evaluation(eval_cfg)
                    actor_pcp, avg_pcp, recall500 = evaluator.evaluate_pcp(
                        preds, recall_threshold=500, alpha=0.5)

                    tb = PrettyTable()
                    tb.field_names = [
                        'Metric', 'Actor 1', 'Actor 2', 'Actor 3', 'Average'
                    ]
                    tb.add_row([
                        'PCP', f'{actor_pcp[0] * 100:.2f}',
                        f'{actor_pcp[1] * 100:.2f}',
                        f'{actor_pcp[2] * 100:.2f}', f'{avg_pcp * 100:.2f}'
                    ])
                    self.logger.info('\n' + tb.get_string())
                    self.logger.info(f'Recall@500mm: {recall500:.4f}')


def train_3d(model,
             optimizer,
             loader,
             epoch: int,
             logger: Union[None, str, logging.Logger],
             output_dir: str,
             clip_max_norm: float,
             print_freq: int,
             n_views: int = 5,
             device: str = 'cuda'):
    """Train one epoch.

    Args:
        model: Model to be trained.
        optimizer:
            Optimizer used in training.
        loader:
            Dataloader.
        epoch (int):
            Current epoch.
        logger (Union[None, str, logging.Logger]):
            Logger for logging. If None, root logger will be selected.
        output_dir (str):
            Path to output folder.
        clip_max_norm (float):
            Gradient clipping.
        print_freq (int):
            Printing frequency during training.
        n_views (int, optional):
            Number of views. Defaults to 5.
        device (str, optional):
            Device name. Defaults to 'cuda'.
    """
    logger = get_logger(logger)
    batch_time = AverageMeter()
    data_time = AverageMeter()
    loss_ce = AverageMeter()
    class_error = AverageMeter()
    loss_per_kp = AverageMeter()
    loss_pose_perprojection = AverageMeter()
    cardinality_error = AverageMeter()

    model.train()

    if model.module.backbone is not None:
        # Comment out this line if you want to train 2D backbone jointly
        model.module.backbone.eval()

    end = time.time()
    for i, (inputs, meta) in enumerate(loader):
        assert len(inputs) == n_views
        inputs = [i.to(device) for i in inputs]
        meta = [{
            k: v.to(device) if isinstance(v, torch.Tensor) else v
            for k, v in t.items()
        } for t in meta]
        data_time.update(time_synchronized() - end)
        end = time_synchronized()

        out, loss_dict, losses = model(views=inputs, meta=meta)

        n_kps = loader.dataset.n_kps
        bs, n_queries = out['pred_logits'].shape[:2]

        src_poses = out['pred_poses']['outputs_coord']. \
            view(bs, n_queries, n_kps, 3)
        src_poses = norm2absolute(src_poses, model.module.grid_size,
                                  model.module.grid_center)
        score = out['pred_logits'][:, :, 1:2].sigmoid()
        score = score.unsqueeze(2).expand(-1, -1, n_kps, -1)

        loss_ce.update(loss_dict['loss_ce'].sum().item())
        class_error.update(loss_dict['class_error'].sum().item())

        loss_per_kp.update(loss_dict['loss_per_kp'].sum().item())

        if 'loss_pose_perprojection' in loss_dict:
            loss_pose_perprojection.update(
                loss_dict['loss_pose_perprojection'].sum().item())

        cardinality_error.update(loss_dict['cardinality_error'].sum().item())

        if losses > 0:
            optimizer.zero_grad()
            losses.backward()
            if clip_max_norm > 0:
                grad_total_norm = torch.nn.utils.clip_grad_norm_(
                    model.parameters(), clip_max_norm)
            else:
                grad_total_norm = get_total_grad_norm(model.parameters(),
                                                      clip_max_norm)

            optimizer.step()

        batch_time.update(time_synchronized() - end)
        end = time_synchronized()

        if i % print_freq == 0 and is_main_process():
            gpu_memory_usage = torch.cuda.memory_allocated(0)
            speed = len(inputs) * inputs[0].size(0) / batch_time.val

            msg = \
                f'Epoch: [{epoch}][{i}/{len(loader)}]\t' \
                f'Time: {batch_time.val:.3f}s ({batch_time.avg:.3f}s)\t' \
                f'Speed: {speed:.1f} samples/s\t' \
                f'Data: {data_time.val:.3f}s ' f'({data_time.avg:.3f}s)\t' \
                f'loss_ce: {loss_ce.val:.7f} ' f'({loss_ce.avg:.7f})\t' \
                f'class_error: {class_error.val:.7f} ' \
                f'({class_error.avg:.7f})\t' \
                f'loss_per_kp: {loss_per_kp.val:.6f} ' \
                f'({loss_per_kp.avg:.6f})\t' \
                f'loss_pose_perprojection: ' \
                f'{loss_pose_perprojection.val:.6f} ' \
                f'({loss_pose_perprojection.avg:.6f})\t' \
                f'cardinality_error: {cardinality_error.val:.6f} ' \
                f'({cardinality_error.avg:.6f})\t' \
                f'Memory {gpu_memory_usage:.1f}\t' \
                f'gradnorm {grad_total_norm:.2f}'

            logger.info(msg)


def validate_3d(model,
                loader,
                logger: Union[None, str, logging.Logger],
                output_dir: str,
                threshold: float,
                print_freq: int,
                n_views: int = 5,
                is_train: bool = False):
    """Evaluate model during training or testing.

    Args:
        model: Model to be evaluated.
        loader: Dataloader.
        logger (Union[None, str, logging.Logger]):
            Logger for logging. If None, root logger will be selected.
        output_dir (str): Path to output folder.
        threshold (float):
            Confidence threshold to filter non-human keypoints.
        print_freq (int): Printing frequency during training.
        n_views (int, optional): Number of views. Defaults to 5.
        is_train (bool, optional):
            True if it is called during trainig. Defaults to False.

    Returns:
        preds(torch.Tensor): Predicted results of all the keypoints.
        keypoints3d(Keypoints): An instance of class keypoints.
    """
    batch_time = AverageMeter()
    data_time = AverageMeter()
    logger = get_logger(logger)

    model.eval()

    preds = []
    keypoints3d = None
    with torch.no_grad():
        end = time.time()
        kps3d_pred = []
        n_max_person = 0
        for i, (inputs, meta) in enumerate(loader):
            data_time.update(time.time() - end)
            assert len(inputs) == n_views
            output = model(views=inputs, meta=meta)

            gt_kps3d = meta[0]['kps3d'].float()
            n_kps = gt_kps3d.shape[2]
            bs, n_queries = output['pred_logits'].shape[:2]

            src_poses = output['pred_poses']['outputs_coord']. \
                view(bs, n_queries, n_kps, 3)
            src_poses = norm2absolute(src_poses, model.module.grid_size,
                                      model.module.grid_center)
            score = output['pred_logits'][:, :, 1:2].sigmoid()
            score = score.unsqueeze(2).expand(-1, -1, n_kps, -1)
            temp = (score > threshold).float() - 1

            pred = torch.cat([src_poses, temp, score], dim=-1)
            pred = pred.detach().cpu().numpy()
            for b in range(pred.shape[0]):
                preds.append(pred[b])

            batch_time.update(time.time() - end)
            end = time.time()
            if (i % print_freq == 0 or i == len(loader) - 1) \
                    and is_main_process():
                gpu_memory_usage = torch.cuda.memory_allocated(0)
                speed = len(inputs) * inputs[0].size(0) / batch_time.val
                msg = f'Test: [{i}/{len(loader)}]\t' \
                      f'Time: {batch_time.val:.3f}s ' \
                      f'({batch_time.avg:.3f}s)\t' \
                      f'Speed: {speed:.1f} samples/s\t' \
                      f'Data: {data_time.val:.3f}s ({data_time.avg:.3f}s)\t' \
                      f'Memory {gpu_memory_usage:.1f}'
                logger.info(msg)

            if not is_train:
                n_person, per_frame_kps3d = convert_result_to_kps(pred)
                n_max_person = max(n_person, n_max_person)
                kps3d_pred.append(per_frame_kps3d)

        if not is_train:
            n_frame = len(kps3d_pred)
            n_kps = n_kps
            kps3d = np.full((n_frame, n_max_person, n_kps, 4), np.nan)

            for frame_idx in range(n_frame):
                per_frame_kps3d = kps3d_pred[frame_idx]
                n_person = len(per_frame_kps3d)
                if n_person > 0:
                    kps3d[frame_idx, :n_person] = per_frame_kps3d

            keypoints3d = Keypoints(kps=kps3d, convention=None)
            kps3d_file = os.path.join(output_dir, 'kps3d.npz')
            if is_main_process():
                logger.info(f'Saving 3D keypoints to: {kps3d_file}')
            keypoints3d.dump(kps3d_file)

    return preds, keypoints3d



# yapf: disable
import numpy as np
import torch
from mmhuman3d.core.conventions.keypoints_mapping import (  # noqa:F401
    KEYPOINTS_FACTORY,
)
from mmhuman3d.core.conventions.keypoints_mapping import \
    convert_kps as convert_kps_mm  # noqa:F401
from mmhuman3d.core.conventions.keypoints_mapping import (  # noqa:F401
    get_keypoint_idx, get_keypoint_idxs_by_part, get_keypoint_num, get_mapping,
)
from typing import List

from xrmocap.data_structure.keypoints import Keypoints
from . import campus, fourdag_19, human_data, panoptic  # noqa:F401
from .paf import ALL_PAF_MAPPING

# yapf: enable
if isinstance(KEYPOINTS_FACTORY, dict):
    KEYPOINTS_FACTORY['campus'] = campus.CAMPUS_KEYPOINTS
    KEYPOINTS_FACTORY['panoptic'] = panoptic.PANOPTIC_KEYPOINTS
    KEYPOINTS_FACTORY['fourdag_19'] = fourdag_19.FOURDAG19_KEYPOINTS


def convert_keypoints(
    keypoints: Keypoints,
    dst: str,
    approximate: bool = False,
    keypoints_factory: dict = KEYPOINTS_FACTORY,
) -> Keypoints:
    """Convert keypoints following the mapping correspondence between src and
    dst keypoints definition.

    Args:
        keypoints (Keypoints):
            An instance of Keypoints class.
        dst (str):
            The name of destination convention.
        approximate (bool, optional):
            Whether approximate mapping is allowed.
            Defaults to False.
        keypoints_factory (dict, optional):
            A dict to store all the keypoint conventions.
            Defaults to KEYPOINTS_FACTORY.

    Returns:
        Keypoints:
            An instance of Keypoints class, whose convention is dst,
            and dtype, device are same as input.
    """
    src = keypoints.get_convention()
    src_arr = keypoints.get_keypoints()
    n_frame, n_person, kps_n, dim = src_arr.shape
    flat_arr = src_arr.reshape(-1, kps_n, dim)
    flat_mask = keypoints.get_mask().reshape(-1, kps_n)

    if isinstance(src_arr, torch.Tensor):

        def new_array_func(shape, value, ref_data, if_uint8):
            if if_uint8:
                dtype = torch.uint8
            else:
                dtype = ref_data.dtype
            if value == 1:
                return torch.ones(
                    size=shape, dtype=dtype, device=ref_data.device)
            elif value == 0:
                return torch.zeros(
                    size=shape, dtype=dtype, device=ref_data.device)
            else:
                raise ValueError

    elif isinstance(src_arr, np.ndarray):

        def new_array_func(shape, value, ref_data, if_uint8):
            if if_uint8:
                dtype = np.uint8
            else:
                dtype = ref_data.dtype
            if value == 1:
                return np.ones(shape=shape)
            elif value == 0:
                return np.zeros(shape=shape, dtype=dtype)
            else:
                raise ValueError

    dst_n_kps = get_keypoint_num(
        convention=dst, keypoints_factory=keypoints_factory)
    dst_idxs, src_idxs, _ = \
        get_mapping(src, dst, approximate, keypoints_factory)
    # multi frame multi person kps
    dst_arr = new_array_func(
        shape=(n_frame * n_person, dst_n_kps, dim),
        value=0,
        ref_data=src_arr,
        if_uint8=False)
    # multi frame multi person mask
    dst_mask = new_array_func(
        shape=(n_frame * n_person, dst_n_kps),
        value=0,
        ref_data=src_arr,
        if_uint8=True)
    # mapping from source
    dst_mask[:, dst_idxs] = flat_mask[:, src_idxs]
    dst_arr[:, dst_idxs, :] = flat_arr[:, src_idxs, :]
    multi_mask = dst_mask.reshape(n_frame, n_person, dst_n_kps)
    multi_arr = dst_arr.reshape(n_frame, n_person, dst_n_kps, dim)
    ret_kps = Keypoints(
        dtype=keypoints.dtype,
        kps=multi_arr,
        mask=multi_mask,
        convention=dst,
        logger=keypoints.logger)
    return ret_kps


def convert_bottom_up_kps_paf(
    kps_paf: List,
    src: str,
    dst: str,
    approximate: bool = False,
    keypoints_factory: dict = KEYPOINTS_FACTORY,
):
    """Convert keypoints and pafs following the mapping correspondence between
    src and dst keypoints definition.

    Args:
        kps_paf (List):
            A list of dict of 2D keypoints and pafs in shape
                 [{'kps':[],'pafs':[]},...]
        src (str):
            The name of source convention.
        dst (str):
            The name of destination convention.
        approximate (bool, optional):
            Whether approximate mapping is allowed.
            Defaults to False.
        keypoints_factory (dict, optional):
            A dict to store all the keypoint conventions.
            Defaults to KEYPOINTS_FACTORY.

    Returns:
        dst_detections (list): the destination keypoints and paf
    """
    n_frame = len(kps_paf)
    dst_n_kps = get_keypoint_num(
        convention=dst, keypoints_factory=keypoints_factory)
    dst_idxs, src_idxs, _ = \
        get_mapping(src, dst, approximate, keypoints_factory)
    paf_mapping = ALL_PAF_MAPPING[src][dst]

    dst_detections = []
    for i in range(n_frame):
        var = {
            'kps': [np.array([]) for j in range(dst_n_kps)],
            'pafs': [np.array([]) for k in range(len(paf_mapping))]
        }
        dst_detections.append(var)
    for frame_id in range(n_frame):
        for i in range(len(dst_idxs)):
            dst_detections[frame_id]['kps'][dst_idxs[i]] = np.array(
                kps_paf[frame_id]['kps'][src_idxs[i]], dtype=np.float32)
        for i in range(len(paf_mapping)):
            if isinstance(paf_mapping[i], list):
                if paf_mapping[i][0] < 0:
                    dst_detections[frame_id]['pafs'][i] = np.array(
                        kps_paf[frame_id]['pafs'][-paf_mapping[i][0]],
                        dtype=np.float32).T
                else:
                    dst_detections[frame_id]['pafs'][i] = np.array(
                        kps_paf[frame_id]['pafs'][paf_mapping[i][0]],
                        dtype=np.float32)
                dst_detections[frame_id]['pafs'][
                    i] = dst_detections[frame_id]['pafs'][i] * (
                        dst_detections[frame_id]['pafs'][i] > 0.1)
                for path_id in paf_mapping[i][1:]:
                    if path_id < 0:
                        arr = np.array(
                            kps_paf[frame_id]['pafs'][-path_id],
                            dtype=np.float32).T
                    else:
                        arr = np.array(
                            kps_paf[frame_id]['pafs'][path_id],
                            dtype=np.float32)
                    dst_detections[frame_id]['pafs'][i] = np.matmul(
                        dst_detections[frame_id]['pafs'][i], arr)
                    dst_detections[frame_id]['pafs'][
                        i] = dst_detections[frame_id]['pafs'][i] * (
                            dst_detections[frame_id]['pafs'][i] > 0.1)
                dst_detections[frame_id]['pafs'][i] = dst_detections[frame_id][
                    'pafs'][i] * len(paf_mapping[i])
            else:
                if paf_mapping[i] < 0:
                    dst_detections[frame_id]['pafs'][i] = np.array(
                        kps_paf[frame_id]['pafs'][-paf_mapping[i]],
                        dtype=np.float32).T
                else:
                    dst_detections[frame_id]['pafs'][i] = np.array(
                        kps_paf[frame_id]['pafs'][paf_mapping[i]],
                        dtype=np.float32)
    return dst_detections


def get_keypoints_factory() -> dict:
    """Get the KEYPOINTS_FACTORY defined in keypoints convention.

    Returns:
        dict:
            KEYPOINTS_FACTORY whose keys are convention
            names and values are keypoints lists.
    """
    return KEYPOINTS_FACTORY


def get_mapping_dict(src: str,
                     dst: str,
                     approximate: bool = False,
                     keypoints_factory: dict = KEYPOINTS_FACTORY) -> dict:
    """Call get_mapping from mmhuman3d and make a dict mapping src index to dst
    index.

    Args:
        src (str):
            The name of source convention.
        dst (str):
            The name of destination convention.
        approximate (bool, optional):
            Whether approximate mapping is allowed.
            Defaults to False.
        keypoints_factory (dict, optional):
            A dict to store all the keypoint conventions.
            Defaults to KEYPOINTS_FACTORY.

    Returns:
        dict:
            A mapping dict whose keys are src indexes
            and values are dst indexes.
    """
    mapping_back = get_mapping(
        src=src,
        dst=dst,
        keypoints_factory=keypoints_factory,
        approximate=approximate)
    inter_to_dst, inter_to_src = mapping_back[:2]
    mapping_dict = {}
    for index in range(len(inter_to_dst)):
        mapping_dict[inter_to_src[index]] = inter_to_dst[index]
    return mapping_dict


def get_keypoint_names(
        convention: str = 'smplx',
        keypoints_factory: dict = KEYPOINTS_FACTORY) -> List[str]:
    """Get names of keypoints of specified convention.

    Args:
        convention (str): data type from keypoints_factory.
        keypoints_factory (dict, optional): A class to store the attributes.
            Defaults to KEYPOINTS_FACTORY.
    Returns:
        List[str]: keypoint names
    """
    keypoints = keypoints_factory[convention]
    return keypoints



import cv2
import logging
import numpy as np
from PIL import Image
from typing import Union

from .base_image_transform import BaseImageTransform


class LoadImageCV2(BaseImageTransform):

    def __init__(self,
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """Load image array from file by cv2.

        Args:
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """
        BaseImageTransform.__init__(self, logger=logger)

    def forward(self, input: str) -> np.ndarray:
        """Forward function of LoadImageCV2.

        Args:
            input (str):
                Path to the image file.

        Returns:
            np.ndarray:
                Image instance defined in PIL.
        """
        img = cv2.imread(input)
        return img


class LoadImagePIL(BaseImageTransform):

    def __init__(self,
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """Load image array from file by PIL.

        Args:
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """
        BaseImageTransform.__init__(self, logger=logger)

    def forward(self, input: str) -> Image:
        """Forward function of LoadImagePIL.

        Args:
            input (str):
                Path to the image file.

        Returns:
            Image:
                Image instance defined in PIL.
        """
        img = Image.open(input)
        return img



import cv2
import h5py
import json
import numpy as np
import tqdm


class SMCReader:

    def __init__(self, file_path, body_model=None):
        """Read SenseMocapFile endswith ".smc".

        Args:
            file_path (str):
                Path to an SMC file.
            body_model (nn.Module or dict):
                Only needed for SMPL transformation to device frame
                if nn.Module: a body_model instance
                if dict: a body_model config
        """
        self.smc = h5py.File(file_path, 'r')
        self.__calibration_dict__ = None
        self.action_id = self.smc.attrs['action_id']
        self.actor_id = self.smc.attrs['actor_id']
        self.datetime_str = self.smc.attrs['datetime_str']  # .decode()
        self.kinect_num_frames = self.smc['Kinect'].attrs['num_frame']
        self.num_kinects = self.smc['Kinect'].attrs['num_device']
        self.kinect_color_resolution = self.get_kinect_color_resolution(0)
        self.kinect_depth_resolution = self.get_kinect_depth_resolution(0)
        self.iphone_exists = 'iPhone' in self.smc.keys()
        self.num_iphones = 1
        if self.iphone_exists:
            self.iphone_num_frames = self.smc['iPhone'].attrs['num_frame']
            self.iphone_color_resolution = \
                self.smc['iPhone'].attrs['color_resolution']  # vertical
            self.iphone_depth_resolution = \
                self.smc['iPhone'].attrs['depth_resolution']  # vertical
        self.keypoint_exists = 'Keypoints3D' in self.smc.keys()

    def get_kinect_color_extrinsics(self, kinect_id, homogeneous=True):
        """Get extrinsics(cam2world) of a kinect RGB camera by kinect id.

        Args:
            kinect_id (int):
                ID of a kinect, starts from 0.
            homogeneous (bool, optional):
                If true, returns rotation and translation in
                one 4x4 matrix. Defaults to True.

        Returns:
            homogeneous is True
                ndarray: A 4x4 matrix of rotation and translation(cam2world).
            homogeneous is False
                dict: A dict of rotation and translation,
                        keys are R and T,
                        each value is an ndarray.
        """
        R = np.asarray(self.calibration_dict[str(kinect_id * 2)]['R']).reshape(
            3, 3)
        T = np.asarray(self.calibration_dict[str(kinect_id *
                                                 2)]['T']).reshape(3)
        if homogeneous:
            extrinsics = np.identity(4, dtype=float)
            extrinsics[:3, :3] = R
            extrinsics[:3, 3] = T
            return extrinsics
        else:
            return {'R': R, 'T': T}

    @property
    def calibration_dict(self):
        """Get the dict of calibration.

        Returns:
            dict:
                A dict of calibrated extrinsics.
        """
        if self.__calibration_dict__ is not None:
            return self.__calibration_dict__
        else:
            return json.loads(self.smc['Extrinsics'][()])

    def get_kinect_depth_extrinsics(self, kinect_id, homogeneous=True):
        """Get extrinsics(cam2world) of a kinect depth camera by kinect id.

        Args:
            kinect_id (int):
                ID of a kinect, starts from 0.
            homogeneous (bool, optional):
                If true, returns rotation and translation in
                one 4x4 matrix. Defaults to True.

        Returns:
            homogeneous is True
                ndarray: A 4x4 matrix of rotation and translation(cam2world).
            homogeneous is False
                dict: A dict of rotation and translation,
                        keys are R and T,
                        each value is an ndarray.
        """
        R = np.asarray(self.calibration_dict[str(kinect_id * 2 +
                                                 1)]['R']).reshape(3, 3)
        T = np.asarray(self.calibration_dict[str(kinect_id * 2 +
                                                 1)]['T']).reshape(3)
        if homogeneous:
            extrinsics = np.identity(4, dtype=float)
            extrinsics[:3, :3] = R
            extrinsics[:3, 3] = T
            return extrinsics
        else:
            return {'R': R, 'T': T}

    def get_kinect_color_intrinsics(self, kinect_id):
        """Get intrinsics of a kinect RGB camera by kinect id.

        Args:
            kinect_id (int):
                ID of a kinect, starts from 0.

        Returns:
            ndarray: A 3x3 matrix.
        """
        kinect_dict = self.smc['Kinect'][str(kinect_id)]
        intrinsics = \
            kinect_dict['Calibration']['Color']['Intrinsics'][()]
        cx, cy, fx, fy = intrinsics[:4]
        intrinsics = \
            np.asarray([[fx, 0, cx], [0, fy, cy], [0, 0, 1]])
        return intrinsics

    def get_kinect_color_resolution(self, kinect_id):
        """Get resolution of a kinect RGB camera by kinect id.

        Args:
            kinect_id (int):
                ID of a kinect, starts from 0.

        Returns:
            ndarray:
                An ndarray of (width, height), shape=[2, ].
        """
        kinect_dict = self.smc['Kinect'][str(kinect_id)]
        resolution = \
            kinect_dict['Calibration']['Color']['Resolution'][()]
        return resolution

    def get_kinect_depth_resolution(self, kinect_id):
        """Get resolution of a kinect depth camera by kinect id.

        Args:
            kinect_id (int):
                ID of a kinect, starts from 0.

        Returns:
            ndarray:
                An ndarray of (width, height), shape=[2, ].
        """
        kinect_dict = self.smc['Kinect'][str(kinect_id)]
        resolution = \
            kinect_dict['Calibration']['Depth']['Resolution'][()]
        return resolution

    def get_kinect_depth_intrinsics(self, kinect_id):
        """Get intrinsics of a kinect depth camera by kinect id.

        Args:
            kinect_id (int):
                ID of a kinect, starts from 0.

        Returns:
            ndarray: A 3x3 matrix.
        """
        kinect_dict = self.smc['Kinect'][str(kinect_id)]
        intrinsics = \
            kinect_dict['Calibration']['Depth']['Intrinsics'][()]
        cx, cy, fx, fy = intrinsics[:4]
        intrinsics = \
            np.asarray([[fx, 0, cx], [0, fy, cy], [0, 0, 1]])
        return intrinsics

    def get_iphone_intrinsics(self, iphone_id=0, frame_id=0, vertical=True):
        """Get intrinsics of an iPhone RGB camera by iPhone id.

        Args:
            iphone_id (int, optional):
                ID of an iPhone, starts from 0.
                Defaults to 0.
            frame_id (int, optional):
                int: frame id of one selected frame
                Defaults to 0.
            vertical (bool, optional):
                iPhone assumes landscape orientation
                if True, convert data to vertical orientation
                Defaults to True.

        Returns:
            ndarray: A 3x3 matrix.
        """
        camera_info = self.smc['iPhone'][str(iphone_id)]['CameraInfo'][str(
            frame_id)]
        camera_info = json.loads(camera_info[()])
        intrinsics = np.asarray(camera_info['cameraIntrinsics']).transpose()

        # Intrinsics have to be adjusted to achieve rotation
        #   1. swapping fx, fy
        #   2. cx -> image height - cy; cy -> cx
        if vertical:
            fx, fy = intrinsics[0, 0], intrinsics[1, 1]
            cx, cy = intrinsics[0, 2], intrinsics[1, 2]
            W, H = self.get_iphone_color_resolution(vertical=False)
            intrinsics = np.eye(3)
            intrinsics[0, 0], intrinsics[1, 1] = fy, fx
            intrinsics[0, 2], intrinsics[1, 2] = H - cy, cx

        return intrinsics

    def get_iphone_extrinsics(self,
                              iphone_id=0,
                              homogeneous=True,
                              vertical=True):
        """Get extrinsics(cam2world) of an iPhone RGB camera by iPhone id.

        Args:
            iphone_id (int, optional):
                ID of an iPhone, starts from 0.
                Defaults to 0.
            homogeneous (bool, optional):
                If true, returns rotation and translation in
                one 4x4 matrix. Defaults to True.
            vertical (bool, optional):
                iPhone assumes landscape orientation
                if True, convert data to vertical orientation
                Defaults to True.

        Returns:
            homogeneous is True
                ndarray: A 4x4 transformation matrix(cam2world).
            homogeneous is False
                dict: A dict of rotation and translation,
                    keys are R and T,
                    each value is an ndarray.
        """
        if iphone_id != 0:
            raise KeyError('Currently only one iPhone.')
        R = np.asarray(self.calibration_dict['iPhone']['R']).reshape(3, 3)
        T = np.asarray(self.calibration_dict['iPhone']['T']).reshape(3)

        # cam2world
        extrinsics = np.identity(4, dtype=float)
        extrinsics[:3, :3] = R
        extrinsics[:3, 3] = T

        # Extrinsics have to be adjusted to achieve rotation
        # A rotation matrix is applied on the extrinsics
        if vertical:
            # 90-degree clockwise rotation around z-axis
            R = np.eye(4)
            R[:2, :2] = np.array([[0, -1], [1, 0]])
            # Note the extrinsics is cam2world
            # world2cam_adjusted = R @ world2cam
            # => cam2world_adjusted = cam2world @ inv(R)
            extrinsics = extrinsics @ np.linalg.inv(R)
            R = extrinsics[:3, :3]
            T = extrinsics[:3, 3]

        if homogeneous:
            return extrinsics
        else:
            return {'R': R, 'T': T}

    def get_iphone_color_resolution(self, iphone_id=0, vertical=True):
        """Get color image resolution of an iPhone RGB camera by iPhone id.

        Args:
            iphone_id (int, optional):
                ID of an iPhone, starts from 0.
                Defaults to 0.
            vertical (bool, optional):
                iPhone assumes landscape orientation
                if True, convert data to vertical orientation
                Defaults to True.

        Returns:
            ndarray:get_iphone_keypoints2d
                An ndarray of (width, height), shape=[2, ].
        """
        if iphone_id != 0:
            raise KeyError('Currently only one iPhone.')
        if vertical:
            W_horizontal, H_horizontal = self.iphone_color_resolution
            W_vertical, H_vertical = H_horizontal, W_horizontal
            return np.array([W_vertical, H_vertical])
        else:
            return self.iphone_color_resolution

    def get_kinect_color(self, kinect_id, frame_id=None, disable_tqdm=True):
        """Get several frames captured by a kinect RGB camera.

        Args:
            kinect_id (int):
                ID of a kinect, starts from 0.
            frame_id (int, list or None, optional):
                int: frame id of one selected frame
                list: a list of frame id
                None: all frames will be returned
                Defaults to None.
            disable_tqdm (bool, optional):
                Whether to disable the entire progressbar wrapper.
                Defaults to True.

        Returns:
            ndarray:
                An ndarray in shape [frame_number, height, width, channels].
        """
        frames = []
        if frame_id is None:
            frame_list = range(self.get_kinect_num_frames())
        elif isinstance(frame_id, list):
            frame_list = frame_id
        elif isinstance(frame_id, int):
            assert frame_id < self.get_kinect_num_frames(),\
                'Index out of range...'
            frame_list = [frame_id]
        else:
            raise TypeError('frame_id should be int, list or None.')
        for i in tqdm.tqdm(frame_list, disable=disable_tqdm):
            frames.append(
                self.__read_color_from_bytes__(
                    self.smc['Kinect'][str(kinect_id)]['Color'][str(i)][()]))
        return np.stack(frames, axis=0)

    def get_kinect_rgbd(self,
                        kinect_id,
                        frame_id,
                        mode='color2depth',
                        threshold=0):
        if mode == 'color2depth':
            mapped_color = \
                self.__map_color_to_depth__(
                    kinect_id, frame_id, threshold=threshold
                )
            depth = self.get_kinect_depth(kinect_id, frame_id)[0]
            return mapped_color, depth
        else:
            print('Model {} is not supported...'.format(mode))

    def get_kinect_depth(self, kinect_id, frame_id=None, disable_tqdm=True):
        """Get several frames captured by a kinect depth camera.

        Args:
            kinect_id (int):
                ID of a kinect, starts from 0.
            frame_id (int, list or None, optional):
                int: frame id of one selected frame
                list: a list of frame id
                None: all frames will be returned
                Defaults to None.
            disable_tqdm (bool, optional):
                Whether to disable the entire progressbar wrapper.
                Defaults to True.

        Returns:
            ndarray:
                An ndarray in shape [frame_number, height, width, channels].
        """
        frames = []
        frame_list = []
        if frame_id is None or type(frame_id) == list:
            frame_list = range(self.get_kinect_num_frames())
            if frame_id:
                frame_list = frame_id
        else:
            assert frame_id < self.get_kinect_num_frames(),\
                'Index out of range...'
            frame_list.append(frame_id)
        for i in tqdm.tqdm(frame_list, disable=disable_tqdm):
            frames.append(
                self.smc['Kinect'][str(kinect_id)]['Depth'][str(i)][()])
        return np.stack(frames, axis=0)

    def __read_color_from_bytes__(self, color_array):
        """Decode an RGB image from an encoded byte array."""
        return cv2.cvtColor(
            cv2.imdecode(color_array, cv2.IMREAD_COLOR), cv2.COLOR_BGR2RGB)

    def get_num_kinect(self):
        """Get the number of Kinect devices.

        Returns:
            int:
                Number of Kinect devices.
        """
        return self.num_kinects

    def get_kinect_num_frames(self):
        """Get the number of frames recorded by one Kinect RGB camera.

        Returns:
            int:
                Number of frames.
        """
        return self.kinect_num_frames

    def get_iphone_num_frames(self):
        """Get the number of frames recorded by one iPhone RGB camera.

        Returns:
            int:
                Number of frames.
        """
        return self.iphone_num_frames

    def get_depth_mask(self, device_id, frame_id):
        return self.smc['Kinect'][str(device_id)]['Mask'][str(frame_id)][()]

    def get_kinect_mask(self, device_id, frame_id):
        kinect_dict = self.smc['Kinect'][str(device_id)]
        return kinect_dict['Mask_k4abt'][str(frame_id)][()]

    def get_num_iphone(self):
        """Get the number of iPhone devices.

        Returns:
            int:
                Number of iPhone devices.
        """
        return self.num_iphones

    def get_iphone_color(self,
                         iphone_id=0,
                         frame_id=None,
                         disable_tqdm=True,
                         vertical=True):
        """Get several frames captured by an iPhone RGB camera.

        Args:
            iphone_id (int):
                ID of an iPhone, starts from 0.
            frame_id (int, list or None, optional):
                int: frame id of one selected frame
                list: a list of frame id
                None: all frames will be returned
                Defaults to None.
            disable_tqdm (bool, optional):
                Whether to disable the entire progressbar wrapper.
                Defaults to True.
            vertical (bool, optional):
                iPhone assumes horizontal orientation
                if True, convert data to vertical orientation
                Defaults to True.

        Returns:
            frames:
                An ndarray in shape [frame_number, height, width, channels].
        """
        frames = []
        if frame_id is None:
            frame_list = range(self.get_iphone_num_frames())
        elif isinstance(frame_id, list):
            frame_list = frame_id
        elif isinstance(frame_id, int):
            assert frame_id < self.get_iphone_num_frames(),\
                'Index out of range...'
            frame_list = [frame_id]
        else:
            raise TypeError('frame_id should be int, list or None.')
        for i in tqdm.tqdm(frame_list, disable=disable_tqdm):
            frame = self.__read_color_from_bytes__(
                self.smc['iPhone'][str(iphone_id)]['Color'][str(i)][()])
            if vertical:
                frame = cv2.rotate(frame, cv2.cv2.ROTATE_90_CLOCKWISE)
            frames.append(frame)
        return np.stack(frames, axis=0)

    def get_iphone_depth(self,
                         iphone_id=0,
                         frame_id=None,
                         disable_tqdm=True,
                         vertical=True):
        """Get several frames captured by an iPhone RGB camera.

        Args:
            iphone_id (int):
                ID of an iPhone, starts from 0.
            frame_id (int, list or None, optional):
                int: frame id of one selected frame
                list: a list of frame id
                None: all frames will be returned
                Defaults to None.
            disable_tqdm (bool, optional):
                Whether to disable the entire progressbar wrapper.
                Defaults to True.
            vertical (bool, optional):
                iPhone assumes horizontal orientation
                if True, convert data to vertical orientation
                Defaults to True.

        Returns:
            frames:
                An ndarray in shape [frame_number, height, width, channels].
        """
        frames = []
        if frame_id is None:
            frame_list = range(self.get_iphone_num_frames())
        elif isinstance(frame_id, list):
            frame_list = frame_id
        elif isinstance(frame_id, int):
            assert frame_id < self.get_iphone_num_frames(),\
                'Index out of range...'
            frame_list = [frame_id]
        else:
            raise TypeError('frame_id should be int, list or None.')
        for i in tqdm.tqdm(frame_list, disable=disable_tqdm):
            frame = self.smc['iPhone'][str(iphone_id)]['Depth'][str(i)][()]
            if vertical:
                frame = cv2.rotate(frame, cv2.cv2.ROTATE_90_CLOCKWISE)
            frames.append(frame)
        return np.stack(frames, axis=0)

    def get_kinect_transformation_depth_to_color(self, device_id):
        """Get transformation matrix from depth to color from a single kinect.

        Args:
            kinect_id (int, optional):
                ID of a Kinect, starts from 0.

        Returns:
            ndarray: A 4x4 transformation matrix.
        """
        return np.linalg.inv(self.get_kinect_color_extrinsics(
            device_id)) @ self.get_kinect_depth_extrinsics(device_id)

    def get_kinect_transformation_color_to_depth(self, device_id):
        """Get transformation matrix from color to depth from a single kinect.

        Args:
            kinect_id (int, optional):
                ID of a Kinect, starts from 0.

        Returns:
            ndarray: A 4x4 transformation matrix.
        """
        return np.linalg.inv(self.get_kinect_depth_extrinsics(
            device_id)) @ self.get_kinect_color_extrinsics(device_id)

    def __map_color_to_depth__(self, device_id, frame_id, threshold=100):
        color_image = self.get_kinect_color(device_id, frame_id)[0]
        depth_image = self.get_kinect_depth(device_id, frame_id)[0]
        color_intrinsic = self.get_kinect_color_intrinsics(device_id)
        depth_intrinsic = self.get_kinect_depth_intrinsics(device_id)

        mask = self.get_depth_mask(device_id, frame_id)

        Td2c = self.get_kinect_transformation_depth_to_color(device_id)

        colidx = np.arange(depth_image.shape[1])
        rowidx = np.arange(depth_image.shape[0])
        colidx_map, rowidx_map = np.meshgrid(colidx, rowidx)
        col_indices = colidx_map[mask >= threshold]
        row_indices = rowidx_map[mask >= threshold]

        homo_padding = \
            np.ones((col_indices.shape[0], 1), dtype=np.float32)
        homo_indices = \
            np.concatenate(
                (col_indices[..., None], row_indices[..., None], homo_padding),
                axis=1
            )

        depth_intrinsic_inv = np.linalg.inv(depth_intrinsic)
        normalized_points = \
            depth_intrinsic_inv[None, ...] @ homo_indices[..., None]

        z_values = (depth_image / 1000)[mask >= threshold]
        valid_points = \
            normalized_points.squeeze() * z_values[..., None]

        R = Td2c[:3, :3]
        T = Td2c[:3, 3]
        valid_points = \
            R[None, ...] @ valid_points[..., None] + T[None, ..., None]
        valid_uvs = \
            color_intrinsic[None, ...] @\
            valid_points / valid_points[:, 2][..., None]
        valid_uvs = np.int32(valid_uvs.squeeze()[..., :2] + 0.5)
        valid_uvs[:, 0] = np.clip(valid_uvs[:, 0], 0, color_image.shape[1] - 1)
        valid_uvs[:, 1] = np.clip(valid_uvs[:, 1], 0, color_image.shape[0] - 1)
        mapped_color = np.ones((depth_image.shape[0], depth_image.shape[1], 3),
                               dtype=np.uint8) * 255
        mapped_color[mask >= threshold] = \
            color_image[valid_uvs[:, 1], valid_uvs[:, 0]]

        if threshold == 1:
            return valid_uvs
        return mapped_color

    def get_kinect_skeleton_3d(self, device_id, frame_id):
        """Get the 3D skeleton key points from a certain kinect.

        Args:
            device_id (int):
                ID of a kinect, starts from 0.

        Returns:
            list:
                A list with 3D keypoints
        """
        kinect_dict = self.smc['Kinect'][str(device_id)]
        return json.loads(kinect_dict['Skeleton_k4abt'][str(frame_id)][()])

    def get_depth_floor(self, device_id: int) -> dict:
        """Get the floor plane defined by a normal vector and a center point
        from a certain kinect.

        Args:
            device_id (int):
                ID of a kinect, starts from 0.

        Raises:
            KeyError:
                Key 'floor' not in ID of a kinect.

        Returns:
            dict:
                A dict with 'center', 'normal' and 'pnum'.
        """
        device_dict = self.calibration_dict[str(device_id * 2 + 1)]
        if 'floor' in device_dict:
            return device_dict['floor']
        else:
            raise KeyError(f'Kinect {device_id} has no floor data.')

    def get_keypoints2d(self, device, device_id, frame_id=None, vertical=True):
        """Get keypoints2d projected from keypoints3d.

        Args:
            device (str):
                Device name, should be Kinect or iPhone.
            device_id (int):
                ID of a device, starts from 0.
            frame_id (int, list or None, optional):
                int: frame id of one selected frame
                list: a list of frame id
                None: all frames will be returned
                Defaults to None.
            vertical (bool, optional):
                Only applicable to iPhone as device
                iPhone assumes horizontal orientation
                if True, convert data to vertical orientation
                Defaults to True.

        Returns:
            Tuple[np.ndarray, np.ndarray]:
                keypoints2d (N, J, 3) and its mask (J, )
        """
        assert device in {
            'Kinect', 'iPhone'
        }, f'Undefined device: {device}, should be "Kinect" or "iPhone"'
        assert device_id >= 0

        kps2d_dict = self.smc['Keypoints2D'][device][str(device_id)]
        keypoints2d = kps2d_dict['keypoints2d'][...]
        keypoints2d_mask = kps2d_dict['keypoints2d_mask'][...]

        if frame_id is None:
            frame_list = range(self.get_keypoints_num_frames())
        elif isinstance(frame_id, list):
            frame_list = frame_id
        elif isinstance(frame_id, int):
            assert frame_id < self.get_keypoints_num_frames(),\
                'Index out of range...'
            frame_list = [frame_id]
        else:
            raise TypeError('frame_id should be int, list or None.')

        keypoints2d = keypoints2d[frame_list, ...]

        if device == 'iPhone' and vertical:
            # rotate keypoints 2D clockwise by 90 degrees
            W, H = self.get_iphone_color_resolution(vertical=False)
            xs, ys, conf = \
                keypoints2d[..., 0], keypoints2d[..., 1], keypoints2d[..., 2]
            xs, ys = H - ys, xs  # horizontal -> vertical
            keypoints2d[..., 0], keypoints2d[..., 1] = xs.copy(), ys.copy()
            keypoints2d[conf == 0.0] = 0.0

        return keypoints2d, keypoints2d_mask

    def get_kinect_keypoints2d(self, device_id, frame_id=None):
        """Get Kinect 2D keypoints.

        Args:
            device_id (int):
                ID of Kinect, starts from 0.
            frame_id (int, list or None, optional):
                int: frame id of one selected frame
                list: a list of frame id
                None: all frames will be returned
                Defaults to None.

        Returns:
            Tuple[np.ndarray, np.ndarray]:
                keypoints2d (N, J, 3) and its mask (J, )
        """
        assert self.num_kinects > device_id >= 0
        return self.get_keypoints2d('Kinect', device_id, frame_id)

    def get_iphone_keypoints2d(self,
                               device_id=0,
                               frame_id=None,
                               vertical=True):
        """Get iPhone 2D keypoints.

        Args:
            device_id (int):
                ID of iPhone, starts from 0.
            frame_id (int, list or None, optional):
                int: frame id of one selected frame
                list: a list of frame id
                None: all frames will be returned
                Defaults to None.
            vertical (bool, optional):
                iPhone assumes horizontal orientation
                if True, convert data to vertical orientation
                Defaults to True.

        Returns:
            Tuple[np.ndarray, np.ndarray]:
                keypoints2d (N, J, 3) and its mask (J, )
        """
        assert device_id >= 0
        return self.get_keypoints2d(
            'iPhone', device_id, frame_id, vertical=vertical)

    def get_color(self,
                  device,
                  device_id,
                  frame_id=None,
                  disable_tqdm=True,
                  vertical=True):
        """Get RGB image(s) from Kinect RGB or iPhone RGB camera.

        Args:
            device (str):
                Device name, should be Kinect or iPhone.
            device_id (int):
                Device ID, starts from 0.
            frame_id (int, list or None, optional):
                int: frame id of one selected frame
                list: a list of frame id
                None: all frames will be returned
                Defaults to None.
            disable_tqdm (bool, optional):
                Whether to disable the entire progressbar wrapper.
                Defaults to True.
            vertical (bool, optional):
                Only applicable to iPhone as device
                iPhone assumes horizontal orientation
                if True, convert data to vertical orientation
                Defaults to True.

        Returns:
            img (ndarray):
                An ndarray in shape [frame_number, height, width, channels].
        """

        assert device in {
            'Kinect', 'iPhone'
        }, f'Undefined device: {device}, should be "Kinect" or "iPhone"'

        if device == 'Kinect':
            img = self.get_kinect_color(device_id, frame_id, disable_tqdm)
        else:
            img = self.get_iphone_color(
                device_id, frame_id, disable_tqdm, vertical=vertical)

        return img

    def get_keypoints_num_frames(self):
        return self.keypoints_num_frames

    def get_keypoints_convention(self):
        return self.keypoints_convention

    def get_keypoints_created_time(self):
        return self.keypoints_created_time

    def get_keypoints3d(self,
                        device=None,
                        device_id=None,
                        frame_id=None,
                        vertical=True):
        """Get keypoints3d (world coordinate) computed by mocap processing
        pipeline.

        Args:
            device (str):
                Device name, should be Kinect or iPhone.
                None: world coordinate
                Defaults to None.
            device_id (int):
                ID of a device, starts from 0.
                None: world coordinate
                Defaults to None
            frame_id (int, list or None, optional):
                int: frame id of one selected frame
                list: a list of frame id
                None: all frames will be returned
                Defaults to None.
            vertical (bool, optional):
                Only applicable to iPhone as device
                iPhone assumes horizontal orientation
                if True, convert data to vertical orientation
                Defaults to True.

        Returns:
            Tuple[np.ndarray, np.ndarray]:
                keypoints3d (N, J, 4) and its mask (J, )
        """
        assert (device is None and device_id is None) or \
            (device is not None and device_id is not None), \
            'device and device_id should be both None or both not None.'
        if device is not None:
            assert device in {
                'Kinect', 'iPhone'
            }, f'Undefined device: {device}, should be "Kinect" or "iPhone"'
        if device_id is not None:
            assert device_id >= 0

        if frame_id is None:
            frame_list = range(self.get_keypoints_num_frames())
        elif isinstance(frame_id, list):
            frame_list = frame_id
        elif isinstance(frame_id, int):
            assert frame_id < self.get_keypoints_num_frames(),\
                'Index out of range...'
            frame_list = [frame_id]
        else:
            raise TypeError('frame_id should be int, list or None.')

        kps3d_dict = self.smc['Keypoints3D']

        # keypoints3d are in world coordinate system
        keypoints3d_world = kps3d_dict['keypoints3d'][...]
        keypoints3d_world = keypoints3d_world[frame_list, ...]
        keypoints3d_mask = kps3d_dict['keypoints3d_mask'][...]

        # return keypoints3d in world coordinate system
        if device is None:
            return keypoints3d_world, keypoints3d_mask

        # return keypoints3d in device coordinate system
        else:
            if device == 'Kinect':
                cam2world = self.get_kinect_color_extrinsics(
                    kinect_id=device_id, homogeneous=True)
            else:
                cam2world = self.get_iphone_extrinsics(
                    iphone_id=device_id, vertical=vertical)

            xyz, conf = keypoints3d_world[..., :3], keypoints3d_world[..., [3]]
            xyz_homogeneous = np.ones([*xyz.shape[:-1], 4])
            xyz_homogeneous[..., :3] = xyz
            world2cam = np.linalg.inv(cam2world)
            keypoints3d = np.einsum('ij,kmj->kmi', world2cam, xyz_homogeneous)
            keypoints3d = np.concatenate([keypoints3d[..., :3], conf], axis=-1)

            return keypoints3d, keypoints3d_mask

    def get_smpl_num_frames(self):
        return self.smpl_num_frames

    def get_smpl_created_time(self):
        return self.smpl_created_time



from mmcv.utils import Registry

from .bottom_up_association_evaluation import BottomUpAssociationEvaluation
from .mvp_evaluation import MVPEvaluation
from .top_down_association_evaluation import TopDownAssociationEvaluation

EVALUATION = Registry('evaluation')

EVALUATION.register_module(
    name='TopDownAssociationEvaluation', module=TopDownAssociationEvaluation)
EVALUATION.register_module(name='MVPEvaluation', module=MVPEvaluation)
EVALUATION.register_module(
    name='BottomUpAssociationEvaluation', module=BottomUpAssociationEvaluation)


def build_evaluation(cfg):
    """Build a matching instance."""
    return EVALUATION.build(cfg)
